import tensorflow as tf
import typeguard as tg
import os

from baconian.common.logging import ConsoleLogger
from baconian.tf.tf_parameters import ParametersWithTensorflowVariable
from baconian.core.core import Basic
from baconian.config.global_config import GlobalConfig


class PlaceholderInput(object):
    @tg.typechecked
    def __init__(self, parameters: ParametersWithTensorflowVariable = None,
                 name_scope=None):
        self.parameters = parameters
        if name_scope:
            self.name_scope = name_scope

    def save(self, global_step, save_path=None, name=None, **kwargs):
        save_path = save_path if save_path else GlobalConfig().DEFAULT_MODEL_CHECKPOINT_PATH
        name = name if name else self.name
        sess = kwargs['sess'] if 'sess' in kwargs else None
        self.parameters.save(save_path=save_path,
                             global_step=global_step,
                             sess=sess,
                             name=name)
        ConsoleLogger().print('info',
                              'model: {}, global step: {}, saved at {}-{}'.format(name, global_step, save_path,
                                                                                  global_step))

    def load(self, path_to_model, model_name, global_step=None, **kwargs):
        sess = kwargs['sess'] if 'sess' in kwargs else None
        self.parameters.load(path_to_model=path_to_model,
                             model_name=model_name,
                             global_step=global_step,
                             sess=sess)
        ConsoleLogger().print('info', 'model: {} loaded from {}'.format(model_name, path_to_model))

    def copy_from(self, obj) -> bool:
        if not isinstance(obj, type(self)):
            raise TypeError('Wrong type of obj %s to be copied, which should be %s' % (type(obj), type(self)))
        self.parameters.copy_from(source_parameter=obj.parameters)
        return True


class MultiPlaceholderInput(object):
    @tg.typechecked
    def __init__(self, sub_placeholder_input_list: list,
                 parameters: ParametersWithTensorflowVariable):
        self._placeholder_input_list = sub_placeholder_input_list
        for param in self._placeholder_input_list:
            assert isinstance(param, dict)
            assert 'attr_name' in param
            assert 'obj' in param and isinstance(param['obj'], PlaceholderInput) and isinstance(param['obj'], Basic)

        self._own_placeholder_input_obj = PlaceholderInput(parameters=parameters)

    def save(self, global_step, save_path, name, **kwargs):
        sess = kwargs['sess'] if 'sess' in kwargs else None
        self._own_placeholder_input_obj.parameters.save(save_path=save_path,
                                                        global_step=global_step,
                                                        sess=sess,
                                                        name=name)
        for param in self._placeholder_input_list:
            param['obj'].save(save_path=os.path.join(save_path, param['attr_name']),
                              global_step=global_step,
                              sess=sess,
                              name=param['obj'].name)
        ConsoleLogger().print('info',
                              'model: {}, global step: {}, saved at {}-{}'.format(name, global_step, save_path,
                                                                                  global_step))

    def load(self, path_to_model, model_name, global_step=None, **kwargs):
        sess = kwargs['sess'] if 'sess' in kwargs else None
        self._own_placeholder_input_obj.parameters.load(
            path_to_model=path_to_model,
            model_name=model_name,
            global_step=global_step,
            sess=sess
        )
        for param in self._placeholder_input_list:
            param['obj'].load(path_to_model=os.path.join(path_to_model, param['attr_name']),
                              global_step=global_step,
                              model_name=param['obj'].name,
                              sess=sess)

        ConsoleLogger().print('info', 'model: {} loaded from {}'.format(model_name, path_to_model))

    def copy_from(self, obj) -> bool:
        if not isinstance(obj, type(self)):
            raise TypeError('Wrong type of obj %s to be copied, which should be %s' % (type(obj), type(self)))
        self._own_placeholder_input_obj.copy_from(obj._own_placeholder_input_obj)
        for self_param, src_param in zip(self._placeholder_input_list, obj._placeholder_input_list):
            self_param['obj'].copy_from(src_param['obj'])
        ConsoleLogger().print('info', 'model: {} copyed from {}'.format(self, obj))
        return True



from baconian.core.core import EnvSpec
from baconian.algo.rl_algo import ModelFreeAlgo, OnPolicyAlgo
from baconian.config.dict_config import DictConfig
import tensorflow as tf
from baconian.algo.distribution.mvn import kl, entropy, log_prob
from baconian.common.sampler.sample_data import TrajectoryData, TransitionData, SampleData
from baconian.tf.tf_parameters import ParametersWithTensorflowVariable
from baconian.config.global_config import GlobalConfig
from baconian.algo.policy.policy import StochasticPolicy
from baconian.algo.value_func import VValueFunction
from baconian.tf.util import *
from baconian.common.misc import *
from baconian.algo.misc import SampleProcessor
from baconian.common.logging import record_return_decorator
from baconian.core.status import register_counter_info_to_status_decorator
from baconian.algo.misc.placeholder_input import MultiPlaceholderInput, PlaceholderInput
from baconian.common.error import *
from baconian.common.data_pre_processing import RunningStandardScaler
from baconian.common.special import *


class PPO(ModelFreeAlgo, OnPolicyAlgo, MultiPlaceholderInput):
    required_key_dict = DictConfig.load_json(file_path=GlobalConfig().DEFAULT_PPO_REQUIRED_KEY_LIST)

    @typechecked
    def __init__(self, env_spec: EnvSpec,
                 stochastic_policy: StochasticPolicy,
                 config_or_config_dict: (DictConfig, dict),
                 value_func: VValueFunction,
                 warm_up_trajectories_number=5,
                 use_time_index_flag=False,
                 name='ppo'):
        ModelFreeAlgo.__init__(self, env_spec=env_spec,
                               name=name,
                               warm_up_trajectories_number=warm_up_trajectories_number)
        self.use_time_index_flag = use_time_index_flag
        self.config = construct_dict_config(config_or_config_dict, self)
        self.policy = stochastic_policy
        self.value_func = value_func
        to_ph_parameter_dict = dict()
        self.trajectory_memory = TrajectoryData(env_spec=env_spec)
        self.transition_data_for_trajectory = TransitionData(env_spec=env_spec)
        self.value_func_train_data_buffer = None
        self.scaler = RunningStandardScaler(dims=self.env_spec.flat_obs_dim)
        if use_time_index_flag:
            scale_last_time_index_mean = self.scaler._mean
            scale_last_time_index_mean[-1] = 0
            scale_last_time_index_var = self.scaler._var
            scale_last_time_index_var[-1] = 1000 * 1000
            self.scaler.set_param(mean=scale_last_time_index_mean, var=scale_last_time_index_var)
        with tf.variable_scope(name):
            self.advantages_ph = tf.placeholder(tf.float32, (None,), 'advantages')
            self.v_func_val_ph = tf.placeholder(tf.float32, (None,), 'val_val_func')
            dist_info_list = self.policy.get_dist_info()
            self.old_dist_tensor = [
                (tf.placeholder(**dict(dtype=dist_info['dtype'],
                                       shape=dist_info['shape'],
                                       name=dist_info['name'])), dist_info['name']) for dist_info in
                dist_info_list
            ]
            self.old_policy = self.policy.make_copy(reuse=False,
                                                    name_scope='old_{}'.format(self.policy.name),
                                                    name='old_{}'.format(self.policy.name),
                                                    distribution_tensors_tuple=tuple(self.old_dist_tensor))
            to_ph_parameter_dict['beta'] = tf.placeholder(tf.float32, (), 'beta')
            to_ph_parameter_dict['eta'] = tf.placeholder(tf.float32, (), 'eta')
            to_ph_parameter_dict['kl_target'] = tf.placeholder(tf.float32, (), 'kl_target')
            to_ph_parameter_dict['lr_multiplier'] = tf.placeholder(tf.float32, (), 'lr_multiplier')

        self.parameters = ParametersWithTensorflowVariable(tf_var_list=[],
                                                           rest_parameters=dict(
                                                               advantages_ph=self.advantages_ph,
                                                               v_func_val_ph=self.v_func_val_ph,
                                                           ),
                                                           to_ph_parameter_dict=to_ph_parameter_dict,
                                                           name='ppo_param',
                                                           save_rest_param_flag=False,
                                                           source_config=self.config,
                                                           require_snapshot=False)
        with tf.variable_scope(name):
            with tf.variable_scope('train'):
                self.kl = tf.reduce_mean(self.old_policy.kl(self.policy))
                self.average_entropy = tf.reduce_mean(self.policy.entropy())
                self.policy_loss, self.policy_optimizer, self.policy_update_op = self._setup_policy_loss()
                self.value_func_loss, self.value_func_optimizer, self.value_func_update_op = self._setup_value_func_loss()
        var_list = get_tf_collection_var_list(
            '{}/train'.format(name)) + self.policy_optimizer.variables() + self.value_func_optimizer.variables()
        self.parameters.set_tf_var_list(tf_var_list=sorted(list(set(var_list)), key=lambda x: x.name))
        MultiPlaceholderInput.__init__(self,
                                       sub_placeholder_input_list=[dict(obj=self.value_func,
                                                                        attr_name='value_func',
                                                                        ),
                                                                   dict(obj=self.policy,
                                                                        attr_name='policy')],
                                       parameters=self.parameters)

    def warm_up(self, trajectory_data: TrajectoryData):
        for traj in trajectory_data.trajectories:
            self.scaler.update_scaler(data=traj.state_set)
        if self.use_time_index_flag:
            scale_last_time_index_mean = self.scaler._mean
            scale_last_time_index_mean[-1] = 0
            scale_last_time_index_var = self.scaler._var
            scale_last_time_index_var[-1] = 1000 * 1000
            self.scaler.set_param(mean=scale_last_time_index_mean, var=scale_last_time_index_var)

    @register_counter_info_to_status_decorator(increment=1, info_key='init', under_status='INITED')
    def init(self, sess=None, source_obj=None):
        self.policy.init()
        self.value_func.init()
        self.parameters.init()
        if source_obj:
            self.copy_from(source_obj)
        super().init()

    @record_return_decorator(which_recorder='self')
    @register_counter_info_to_status_decorator(increment=1, info_key='train', under_status='TRAIN')
    def train(self, trajectory_data: TrajectoryData = None, train_iter=None, sess=None) -> dict:
        super(PPO, self).train()
        if trajectory_data is None:
            trajectory_data = self.trajectory_memory
        if len(trajectory_data) == 0:
            raise MemoryBufferLessThanBatchSizeError('not enough trajectory data')
        for i, traj in enumerate(trajectory_data.trajectories):
            trajectory_data.trajectories[i].append_new_set(name='state_set',
                                                           shape=self.env_spec.obs_shape,
                                                           data_set=np.reshape(
                                                               np.array(self.scaler.process(np.array(traj.state_set))),
                                                               [-1] + list(self.env_spec.obs_shape)))
            trajectory_data.trajectories[i].append_new_set(name='new_state_set',
                                                           shape=self.env_spec.obs_shape,
                                                           data_set=np.reshape(
                                                               np.array(
                                                                   self.scaler.process(np.array(traj.new_state_set))),
                                                               [-1] + list(self.env_spec.obs_shape)))

        tf_sess = sess if sess else tf.get_default_session()
        SampleProcessor.add_estimated_v_value(trajectory_data, value_func=self.value_func)
        SampleProcessor.add_discount_sum_reward(trajectory_data,
                                                gamma=self.parameters('gamma'))
        SampleProcessor.add_gae(trajectory_data,
                                gamma=self.parameters('gamma'),
                                name='advantage_set',
                                lam=self.parameters('lam'),
                                value_func=self.value_func)
        trajectory_data = SampleProcessor.normalization(trajectory_data, key='advantage_set')
        policy_res_dict = self._update_policy(
            state_set=np.concatenate([t('state_set') for t in trajectory_data.trajectories], axis=0),
            action_set=np.concatenate([t('action_set') for t in trajectory_data.trajectories], axis=0),
            advantage_set=np.concatenate([t('advantage_set') for t in trajectory_data.trajectories], axis=0),
            train_iter=train_iter if train_iter else self.parameters(
                'policy_train_iter'),
            sess=tf_sess)
        value_func_res_dict = self._update_value_func(
            state_set=np.concatenate([t('state_set') for t in trajectory_data.trajectories], axis=0),
            discount_set=np.concatenate([t('discount_set') for t in trajectory_data.trajectories], axis=0),
            train_iter=train_iter if train_iter else self.parameters(
                'value_func_train_iter'),
            sess=tf_sess)
        trajectory_data.reset()
        self.trajectory_memory.reset()
        return {
            **policy_res_dict,
            **value_func_res_dict
        }

    @register_counter_info_to_status_decorator(increment=1, info_key='test', under_status='TEST')
    def test(self, *arg, **kwargs) -> dict:
        return super().test(*arg, **kwargs)

    @register_counter_info_to_status_decorator(increment=1, info_key='predict')
    def predict(self, obs: np.ndarray, sess=None, batch_flag: bool = False):
        tf_sess = sess if sess else tf.get_default_session()
        ac = self.policy.forward(obs=self.scaler.process(data=make_batch(obs, original_shape=self.env_spec.obs_shape)),
                                 sess=tf_sess,
                                 feed_dict=self.parameters.return_tf_parameter_feed_dict())
        return ac

    def append_to_memory(self, samples: TrajectoryData):
        # todo how to make sure the data's time sequential
        obs_list = samples.trajectories[0].state_set
        for i in range(1, len(samples.trajectories)):
            obs_list = np.array(np.concatenate([obs_list, samples.trajectories[i].state_set], axis=0))
        self.trajectory_memory.union(samples)
        self.scaler.update_scaler(data=np.array(obs_list))
        if self.use_time_index_flag:
            scale_last_time_index_mean = self.scaler._mean
            scale_last_time_index_mean[-1] = 0
            scale_last_time_index_var = self.scaler._var
            scale_last_time_index_var[-1] = 1000 * 1000
            self.scaler.set_param(mean=scale_last_time_index_mean, var=scale_last_time_index_var)

    @record_return_decorator(which_recorder='self')
    def save(self, global_step, save_path=None, name=None, **kwargs):
        save_path = save_path if save_path else GlobalConfig().DEFAULT_MODEL_CHECKPOINT_PATH
        name = name if name else self.name
        MultiPlaceholderInput.save(self, save_path=save_path, global_step=global_step, name=name, **kwargs)
        return dict(check_point_save_path=save_path, check_point_save_global_step=global_step,
                    check_point_save_name=name)

    @record_return_decorator(which_recorder='self')
    def load(self, path_to_model, model_name, global_step=None, **kwargs):
        MultiPlaceholderInput.load(self, path_to_model, model_name, global_step, **kwargs)
        return dict(check_point_load_path=path_to_model, check_point_load_global_step=global_step,
                    check_point_load_name=model_name)

    def _setup_policy_loss(self):
        """
        Code clip from pat-cody
        Three loss terms:
            1) standard policy gradient
            2) D_KL(pi_old || pi_new)
            3) Hinge loss on [D_KL - kl_targ]^2

        See: https://arxiv.org/pdf/1707.02286.pdf
        """

        if self.parameters('clipping_range') is not None:
            pg_ratio = tf.exp(self.policy.log_prob() - self.old_policy.log_prob())
            clipped_pg_ratio = tf.clip_by_value(pg_ratio, 1 - self.parameters('clipping_range')[0],
                                                1 + self.parameters('clipping_range')[1])
            surrogate_loss = tf.minimum(self.advantages_ph * pg_ratio,
                                        self.advantages_ph * clipped_pg_ratio)
            loss = -tf.reduce_mean(surrogate_loss)
        else:
            loss1 = -tf.reduce_mean(self.advantages_ph *
                                    tf.exp(self.policy.log_prob() - self.old_policy.log_prob()))
            loss2 = tf.reduce_mean(self.parameters('beta') * self.kl)
            loss3 = self.parameters('eta') * tf.square(
                tf.maximum(0.0, self.kl - 2.0 * self.parameters('kl_target')))
            loss = loss1 + loss2 + loss3
            self.loss1 = loss1
            self.loss2 = loss2
            self.loss3 = loss3
        if isinstance(self.policy, PlaceholderInput):
            reg_list = tf.get_collection(tf.GraphKeys.REGULARIZATION_LOSSES, scope=self.policy.name_scope)
            if len(reg_list) > 0:
                reg_loss = tf.reduce_sum(reg_list)
                loss += reg_loss

        optimizer = tf.train.AdamOptimizer(
            learning_rate=self.parameters('policy_lr') * self.parameters('lr_multiplier'))
        train_op = optimizer.minimize(loss, var_list=self.policy.parameters('tf_var_list'))
        return loss, optimizer, train_op

    def _setup_value_func_loss(self):
        # todo update the value_func design
        loss = tf.reduce_mean(tf.square(tf.squeeze(self.value_func.v_tensor) - self.v_func_val_ph))
        reg_loss = tf.get_collection(tf.GraphKeys.REGULARIZATION_LOSSES, scope=self.value_func.name_scope)
        if len(reg_loss) > 0:
            loss += tf.reduce_sum(reg_loss)
        optimizer = tf.train.AdamOptimizer(self.parameters('value_func_lr'))
        train_op = optimizer.minimize(loss, var_list=self.value_func.parameters('tf_var_list'))
        return loss, optimizer, train_op

    def _update_policy(self, state_set, action_set, advantage_set, train_iter, sess):
        old_policy_feed_dict = dict()
        res = sess.run([getattr(self.policy, tensor[1]) for tensor in self.old_dist_tensor],
                       feed_dict={
                           self.policy.parameters('state_input'): state_set,
                           self.policy.parameters('action_input'): action_set,
                           **self.parameters.return_tf_parameter_feed_dict()
                       })

        for tensor, val in zip(self.old_dist_tensor, res):
            old_policy_feed_dict[tensor[0]] = val

        feed_dict = {
            self.policy.parameters('action_input'): action_set,
            self.old_policy.parameters('action_input'): action_set,
            self.policy.parameters('state_input'): state_set,
            self.advantages_ph: advantage_set,
            **self.parameters.return_tf_parameter_feed_dict(),
            **old_policy_feed_dict
        }
        average_loss, average_kl, average_entropy = 0.0, 0.0, 0.0
        total_epoch = 0
        kl = None
        for i in range(train_iter):
            _ = sess.run(self.policy_update_op, feed_dict=feed_dict)
            loss, kl, entropy = sess.run(
                [self.policy_loss, self.kl, self.average_entropy],
                feed_dict=feed_dict)
            average_loss += loss
            average_kl += kl
            average_entropy += entropy
            total_epoch = i + 1
            if kl > self.parameters('kl_target', require_true_value=True) * 4:
                # early stopping if D_KL diverges badly
                break
        average_loss, average_kl, average_entropy = average_loss / total_epoch, average_kl / total_epoch, average_entropy / total_epoch

        if kl > self.parameters('kl_target', require_true_value=True) * 2:  # servo beta to reach D_KL target
            self.parameters.set(key='beta',
                                new_val=np.minimum(35, 1.5 * self.parameters('beta', require_true_value=True)))
            if self.parameters('beta', require_true_value=True) > 30 and self.parameters('lr_multiplier',
                                                                                         require_true_value=True) > 0.1:
                self.parameters.set(key='lr_multiplier',
                                    new_val=self.parameters('lr_multiplier', require_true_value=True) / 1.5)
        elif kl < self.parameters('kl_target', require_true_value=True) / 2:
            self.parameters.set(key='beta',
                                new_val=np.maximum(1 / 35, self.parameters('beta', require_true_value=True) / 1.5))

            if self.parameters('beta', require_true_value=True) < (1 / 30) and self.parameters('lr_multiplier',
                                                                                               require_true_value=True) < 10:
                self.parameters.set(key='lr_multiplier',
                                    new_val=self.parameters('lr_multiplier', require_true_value=True) * 1.5)
        return dict(
            policy_average_loss=average_loss,
            policy_average_kl=average_kl,
            policy_average_entropy=average_entropy,
            policy_total_train_epoch=total_epoch
        )

    def _update_value_func(self, state_set, discount_set, train_iter, sess):
        y_hat = self.value_func.forward(obs=state_set).squeeze()
        old_exp_var = 1 - np.var(discount_set - y_hat) / np.var(discount_set)

        if self.value_func_train_data_buffer is None:
            self.value_func_train_data_buffer = (state_set, discount_set)
        else:
            self.value_func_train_data_buffer = (
                np.concatenate([state_set, self.value_func_train_data_buffer[0]], axis=0),
                np.concatenate([discount_set, self.value_func_train_data_buffer[1]], axis=0))

        state_set_all, discount_set_all = self.value_func_train_data_buffer

        param_dict = self.parameters.return_tf_parameter_feed_dict()

        for i in range(train_iter):
            random_index = np.random.choice(np.arange(len(state_set_all)), len(state_set_all))
            state_set_all = state_set_all[random_index]
            discount_set_all = discount_set_all[random_index]
            for index in range(0,
                               len(state_set_all) - self.parameters('value_func_train_batch_size'),
                               self.parameters('value_func_train_batch_size')):
                state = np.array(state_set_all[
                                 index: index + self.parameters(
                                     'value_func_train_batch_size')])
                discount = discount_set_all[index: index + self.parameters(
                    'value_func_train_batch_size')]
                loss, _ = sess.run([self.value_func_loss, self.value_func_update_op],
                                   options=tf.RunOptions(report_tensor_allocations_upon_oom=True),
                                   feed_dict={
                                       self.value_func.state_input: state,
                                       self.v_func_val_ph: discount,
                                       **param_dict
                                   })
        y_hat = self.value_func.forward(obs=state_set).squeeze()
        loss = np.mean(np.square(y_hat - discount_set))
        exp_var = 1 - np.var(discount_set - y_hat) / np.var(discount_set)
        return dict(
            value_func_loss=loss,
            value_func_policy_exp_var=exp_var,
            value_func_policy_old_exp_var=old_exp_var
        )



"""
A scikit-learn liked module for handle the data pre-processing including normalization, standardization,
"""

import numpy as np
from baconian.common.error import *


class DataScaler(object):
    def __init__(self, dims):
        self.data_dims = dims

    def _check_scaler(self, scaler) -> bool:
        if len(scaler.shape) != 1 or scaler.shape[0] != self.data_dims:
            return False
        else:
            return True

    def _compute_stat_of_batch_data(self, data):
        data = np.array(data)
        if self._check_data(data) is False:
            raise ShapeNotCompatibleError("data shape is not compatible")
        else:
            return np.min(data, axis=0), np.max(data, axis=0), np.mean(data, axis=0), np.var(data, axis=0)

    def _check_data(self, data) -> bool:
        if len(data.shape) != 2 or data.shape[1] != self.data_dims:
            return False
        else:
            return True

    def process(self, data):
        raise NotImplementedError

    def inverse_process(self, data):
        raise NotImplementedError

    def update_scaler(self, data):
        pass


class IdenticalDataScaler(DataScaler):
    def __init__(self, dims):
        self.data_dims = dims

    def process(self, data):
        return data

    def inverse_process(self, data):
        return data


class MinMaxScaler(DataScaler):
    def __init__(self, dims: int, desired_range: tuple = None):
        super().__init__(dims)
        self._min = np.zeros([dims])
        self._max = np.ones([dims])

        if desired_range is None:
            self._desired_range = (np.zeros(dims), np.ones(dims))
        else:
            if len(desired_range) != 2 or self._check_scaler(np.array(desired_range[0])) is False or self._check_scaler(
                    np.array(desired_range[1])) is False:
                raise ShapeNotCompatibleError("desired value dims is not compatible with dims")
            self._desired_range = (np.array(desired_range[0]), np.array(desired_range[1]))

    def process(self, data):
        return (data - self._min) / (self._max - self._min) * (self._desired_range[1] - self._desired_range[0]) + \
               self._desired_range[0]

    def get_param(self):
        return dict(min=self._min.tolist(),
                    max=self._max.tolist(),
                    desired_range=np.array(self._desired_range).tolist())

    def set_param(self, min=None, max=None, desired_range=None):
        if self._check_scaler(min):
            self._min = min
        else:
            raise ShapeNotCompatibleError('the shape of min/max range is not as same as shape')
        if self._check_scaler(max):
            self._max = max
        else:
            raise ShapeNotCompatibleError('the shape of min/max range is not as same as shape')
        if len(desired_range) != 2 and self._check_scaler(np.array(desired_range[0])) is False or self._check_scaler(
                np.array(desired_range[1])) is False:
            raise ShapeNotCompatibleError("desired value dims is not compatible with dims")
        else:
            self._desired_range = (np.array(desired_range[0]), np.array(desired_range[1]))

    def inverse_process(self, data):
        if np.greater(np.array(data), self._desired_range[1]).any() or np.less(data, self._desired_range[0]).any():
            raise WrongValueRangeError('data for inverse process not in the range {} {}'.format(self._desired_range[0],
                                                                                                self._desired_range[1]))
        return (np.array(data) - self._desired_range[0]) / (self._desired_range[1] - self._desired_range[0]) * \
               (self._max - self._min) + self._min


class RunningMinMaxScaler(MinMaxScaler):
    """
    A scaler with running mean and max across all data updated into the scaler and scale the data to a desired range
    """

    def __init__(self, dims: int, desired_range: tuple = None, init_data: np.ndarray = None,
                 init_min: np.ndarray = None,
                 init_max: np.ndarray = None):
        super().__init__(dims=dims, desired_range=desired_range)
        if init_max is not None and init_max is not None:
            self._min = np.array(init_min)
            self._max = np.array(init_max)
        elif init_data is not None:
            self._min, self._max, _, _ = self._compute_stat_of_batch_data(init_data)
        if self._check_scaler(self._min) is False or self._check_scaler(self._max) is False:
            raise ShapeNotCompatibleError('the shape of min/max range is not as same as shape')

    def update_scaler(self, data):
        if len(np.array(data)) == 0:
            return
        if self._min is not None:
            self._min = np.minimum(np.min(data, axis=0), self._min)
        else:
            self._min = np.min(data, axis=0)
        if self._max is not None:
            self._max = np.maximum(np.max(data, axis=0), self._max)
        else:
            self._max = np.max(data, axis=0)


class BatchMinMaxScaler(MinMaxScaler):

    def process(self, data):
        if self._check_data(data) is False:
            raise ShapeNotCompatibleError("data is compatible with scaler, make sure only scale the box type data")
        self._min, self._max, _, _ = self._compute_stat_of_batch_data(data)
        return super().process(data)


class StandardScaler(DataScaler):
    def __init__(self, dims: int):
        super().__init__(dims)
        self._var = np.ones([dims])
        self._mean = np.zeros([dims])
        self._data_count = 0
        self._epsilon = 0.01

    def process(self, data):
        return (np.array(data) - self._mean) / (np.sqrt(self._var) + self._epsilon)

    def inverse_process(self, data):
        return np.array(data) * (np.sqrt(self._var) + self._epsilon) + self._mean

    def get_param(self):
        return dict(mean=self._mean.tolist(),
                    var=self._var.tolist())


class BatchStandardScaler(StandardScaler):
    def process(self, data):
        if self._check_data(data) is False:
            raise ShapeNotCompatibleError("data is compatible with scaler")
        _, _, self._mean, self._var = self._compute_stat_of_batch_data(data)
        return super().process(data)


class RunningStandardScaler(StandardScaler):
    """
    A scaler with running mean and variance across all data passed into the scaler and scale the data with zero mean and
    unit variance.
    """

    def __init__(self, dims: int, init_data: np.ndarray = None,
                 init_mean: np.ndarray = None,
                 init_var: np.ndarray = None,
                 init_mean_var_data_count=None):
        super().__init__(dims)

        if init_mean is not None and init_var is not None:
            self._mean = init_mean
            self._var = init_var
            self._data_count = init_mean_var_data_count
        elif init_data is not None:
            _, _, self._mean, self._var = self._compute_stat_of_batch_data(init_data)
            self._data_count = np.array(init_data).shape[0]

    def update_scaler(self, data):
        if len(np.array(data)) == 0:
            return
        if self._mean is None or self._var is None:
            _, _, self._mean, self._var = self._compute_stat_of_batch_data(data)
            self._data_count = data.shape[0]
        else:
            n = data.shape[0]
            new_data_var = np.var(data, axis=0)
            new_data_mean = np.mean(data, axis=0)
            new_data_mean_sq = np.square(new_data_mean)
            new_means = ((self._mean * self._data_count) + (new_data_mean * n)) / (self._data_count + n)
            self._var = (((self._data_count * (self._var + np.square(self._mean))) +
                          (n * (new_data_var + new_data_mean_sq))) / (self._data_count + n) -
                         np.square(new_means))
            self._var = np.maximum(0.0, self._var)
            self._mean = new_means
            self._data_count += n

    def set_param(self, mean=None, var=None):
        assert np.array(mean).shape == self._mean.shape, 'new mean shape is changed'
        assert np.array(var).shape == self._var.shape, 'new var shape is changed'
        self._mean = mean
        self._var = var



"""
An example showing the Model-ensemble method in
Kurutach, Thanard, et al. "Model-ensemble trust-region policy optimization." arXiv preprint arXiv:1802.10592 (2018).

Here we use Model-ensemble with DDPG instead of TRPO on Pendulum-v1, also resue the Dyna flow to show the flexibility of Baconian modules.

"""

from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.algo.value_func.mlp_q_value import MLPQValueFunction
from baconian.algo.ddpg import DDPG
from baconian.algo.policy import DeterministicMLPPolicy
from baconian.core.agent import Agent
from baconian.algo.misc import EpsilonGreedy
from baconian.core.experiment import Experiment
from baconian.config.global_config import GlobalConfig
from baconian.core.status import get_global_status_collect
from baconian.common.schedules import PeriodicalEventSchedule
from baconian.algo.dynamics.mlp_dynamics_model import ContinuousMLPGlobalDynamicsModel
from baconian.algo.model_ensemble import ModelEnsembleAlgo
from baconian.envs.envs_reward_func import PendulumRewardFunc
from baconian.algo.dynamics.terminal_func.terminal_func import FixedEpisodeLengthTerminalFunc
from baconian.core.flow.dyna_flow import create_dyna_flow
from baconian.common.data_pre_processing import RunningStandardScaler
from baconian.core.ensemble import ModelEnsemble


def task_fn():
    env = make('Pendulum-v0')
    name = 'demo_exp'
    env_spec = EnvSpec(obs_space=env.observation_space,
                       action_space=env.action_space)

    mlp_q = MLPQValueFunction(env_spec=env_spec,
                              name_scope=name + '_mlp_q',
                              name=name + '_mlp_q',
                              mlp_config=[
                                  {
                                      "ACT": "RELU",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "1",
                                      "N_UNITS": 16,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  },
                                  {
                                      "ACT": "LINEAR",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "OUPTUT",
                                      "N_UNITS": 1,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  }
                              ])
    policy = DeterministicMLPPolicy(env_spec=env_spec,
                                    name_scope=name + '_mlp_policy',
                                    name=name + '_mlp_policy',
                                    mlp_config=[
                                        {
                                            "ACT": "RELU",
                                            "B_INIT_VALUE": 0.0,
                                            "NAME": "1",
                                            "N_UNITS": 16,
                                            "TYPE": "DENSE",
                                            "W_NORMAL_STDDEV": 0.03
                                        },
                                        {
                                            "ACT": "LINEAR",
                                            "B_INIT_VALUE": 0.0,
                                            "NAME": "OUPTUT",
                                            "N_UNITS": env_spec.flat_action_dim,
                                            "TYPE": "DENSE",
                                            "W_NORMAL_STDDEV": 0.03
                                        }
                                    ],
                                    reuse=False)

    ddpg = DDPG(
        env_spec=env_spec,
        config_or_config_dict={
            "REPLAY_BUFFER_SIZE": 10000,
            "GAMMA": 0.999,
            "CRITIC_LEARNING_RATE": 0.001,
            "ACTOR_LEARNING_RATE": 0.001,
            "DECAY": 0.5,
            "BATCH_SIZE": 50,
            "TRAIN_ITERATION": 1,
            "critic_clip_norm": 0.1,
            "actor_clip_norm": 0.1,
        },
        value_func=mlp_q,
        policy=policy,
        name=name + '_ddpg',
        replay_buffer=None
    )
    mlp_dyna_list = []
    for i in range(10):
        mlp_dyna = ContinuousMLPGlobalDynamicsModel(
            env_spec=env_spec,
            name_scope=name + '_mlp_dyna_{}'.format(i),
            name=name + '_mlp_dyna_{}'.format(i),
            learning_rate=0.01,
            state_input_scaler=RunningStandardScaler(dims=env_spec.flat_obs_dim),
            action_input_scaler=RunningStandardScaler(dims=env_spec.flat_action_dim),
            output_delta_state_scaler=RunningStandardScaler(dims=env_spec.flat_obs_dim),
            mlp_config=[
                {
                    "ACT": "RELU",
                    "B_INIT_VALUE": 0.0,
                    "NAME": "1",
                    "L1_NORM": 0.0,
                    "L2_NORM": 0.0,
                    "N_UNITS": 16,
                    "TYPE": "DENSE",
                    "W_NORMAL_STDDEV": 0.03
                },
                {
                    "ACT": "LINEAR",
                    "B_INIT_VALUE": 0.0,
                    "NAME": "OUPTUT",
                    "L1_NORM": 0.0,
                    "L2_NORM": 0.0,
                    "N_UNITS": env_spec.flat_obs_dim,
                    "TYPE": "DENSE",
                    "W_NORMAL_STDDEV": 0.03
                }
            ])
        mlp_dyna_list.append(mlp_dyna)
    dyna_ensemble_model = ModelEnsemble(n_models=10, model=mlp_dyna_list, prediction_type='random', env_spec=env_spec)
    algo = ModelEnsembleAlgo(
        env_spec=env_spec,
        model_free_algo=ddpg,
        dynamics_model=dyna_ensemble_model,
        config_or_config_dict=dict(
            dynamics_model_train_iter=10,
            model_free_algo_train_iter=10,
            validation_trajectory_count=2,
        )
    )
    # For examples only, we use random reward function and terminal function with fixed episode length.
    algo.set_terminal_reward_function_for_dynamics_env(
        terminal_func=FixedEpisodeLengthTerminalFunc(max_step_length=env.unwrapped._max_episode_steps,
                                                     step_count_fn=algo.dynamics_env.total_step_count_fn),
        reward_func=PendulumRewardFunc())
    agent = Agent(env=env, env_spec=env_spec,
                  algo=algo,
                  algo_saving_scheduler=PeriodicalEventSchedule(
                      t_fn=lambda: get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
                      trigger_every_step=200,
                      after_t=10),
                  name=name + '_agent',
                  exploration_strategy=EpsilonGreedy(action_space=env_spec.action_space,
                                                     init_random_prob=0.5))

    # we can easily reuse the dyna training flow to implement the Model-ensemble training flow.
    flow = create_dyna_flow(
        train_algo_func=(agent.train, (), dict(state='state_agent_training')),
        train_algo_from_synthesized_data_func=(agent.train, (), dict(state='state_agent_training')),
        train_dynamics_func=(agent.train, (), dict(state='state_dynamics_training')),
        test_algo_func=(agent.test, (), dict(sample_count=10)),
        test_dynamics_func=(agent.algo.test_dynamics, (), dict(sample_count=10, env=env)),
        sample_from_real_env_func=(agent.sample, (), dict(sample_count=10,
                                                          env=agent.env,
                                                          store_flag=True)),
        sample_from_dynamics_env_func=(agent.sample, (), dict(sample_count=10,
                                                              env=agent.algo.dynamics_env,
                                                              store_flag=True)),
        # set this to large enough so agent only use data from dynamics env.
        train_algo_every_real_sample_count_by_data_from_real_env=100,
        train_algo_every_real_sample_count_by_data_from_dynamics_env=100,
        test_algo_every_real_sample_count=100,
        test_dynamics_every_real_sample_count=100,
        train_dynamics_ever_real_sample_count=100,
        start_train_algo_after_sample_count=1,
        start_train_dynamics_after_sample_count=1,
        start_test_algo_after_sample_count=1,
        start_test_dynamics_after_sample_count=1,
        warm_up_dynamics_samples=100
    )

    experiment = Experiment(
        tuner=None,
        env=env,
        agent=agent,
        flow=flow,
        name=name + '_exp'
    )
    experiment.run()


from baconian.core.experiment_runner import *

GlobalConfig().set('DEFAULT_LOG_PATH', './log_path')
GlobalConfig().set('DEFAULT_EXPERIMENT_END_POINT', dict(TOTAL_AGENT_TRAIN_SAMPLE_COUNT=400))

single_exp_runner(task_fn, del_if_log_path_existed=True)



from baconian.test.tests.set_up.setup import TestWithAll
from baconian.config.global_config import GlobalConfig
import glob


class TestPlaceholderInput(TestWithAll):
    def test_tf_param(self):
        a, _ = self.create_ph('test')
        for i in range(5):
            a.save(save_path=GlobalConfig().DEFAULT_LOG_PATH + '/test_placehoder_input',
                   global_step=i,
                   name='a')
        file = glob.glob(GlobalConfig().DEFAULT_LOG_PATH + '/test_placehoder_input/a*.meta')
        self.assertTrue(len(file) == 5)
        b, _ = self.create_ph('b')
        b.copy_from(obj=a)
        self.assert_var_list_equal(a.parameters('tf_var_list'),
                                   b.parameters('tf_var_list'))

        a.parameters.init()
        self.assert_var_list_at_least_not_equal(a.parameters('tf_var_list'),
                                                b.parameters('tf_var_list'))

        a.load(path_to_model=GlobalConfig().DEFAULT_LOG_PATH + '/test_placehoder_input',
               global_step=4,
               model_name='a')

        self.assert_var_list_equal(a.parameters('tf_var_list'),
                                   b.parameters('tf_var_list'))

    def test_save_load_with_dqn(self):
        dqn, locals = self.create_dqn()
        dqn.init()
        for i in range(5):
            dqn.save(save_path=GlobalConfig().DEFAULT_LOG_PATH + '/test_placehoder_input', global_step=i, name='dqn')
        file = glob.glob(GlobalConfig().DEFAULT_LOG_PATH + '/test_placehoder_input/dqn*.meta')
        self.assertTrue(len(file) == 5)
        dqn2, _ = self.create_dqn(name='dqn_2')
        dqn2.copy_from(dqn)

        self.assert_var_list_equal(dqn.parameters('tf_var_list'), dqn2.parameters('tf_var_list'))
        self.assert_var_list_equal(dqn.q_value_func.parameters('tf_var_list'),
                                   dqn2.q_value_func.parameters('tf_var_list'))
        self.assert_var_list_equal(dqn.target_q_value_func.parameters('tf_var_list'),
                                   dqn2.target_q_value_func.parameters('tf_var_list'))

        dqn.init()
        self.assert_var_list_at_least_not_equal(dqn.q_value_func.parameters('tf_var_list'),
                                                dqn2.q_value_func.parameters('tf_var_list'))
        self.assert_var_list_at_least_not_equal(dqn.target_q_value_func.parameters('tf_var_list'),
                                                dqn2.target_q_value_func.parameters('tf_var_list'))
        dqn.load(path_to_model=GlobalConfig().DEFAULT_LOG_PATH + '/test_placehoder_input', global_step=4,
                 model_name='dqn')

        self.assert_var_list_equal(dqn.parameters('tf_var_list'), dqn2.parameters('tf_var_list'))
        self.assert_var_list_equal(dqn.q_value_func.parameters('tf_var_list'),
                                   dqn2.q_value_func.parameters('tf_var_list'))
        self.assert_var_list_equal(dqn.target_q_value_func.parameters('tf_var_list'),
                                   dqn2.target_q_value_func.parameters('tf_var_list'))



from baconian.algo.policy.policy import StochasticPolicy
from baconian.core.core import EnvSpec
import overrides
import tensorflow as tf
from baconian.tf.mlp import MLP
from baconian.tf.tf_parameters import ParametersWithTensorflowVariable
from baconian.common.special import *
from baconian.tf.util import *
from baconian.algo.utils import _get_copy_arg_with_tf_reuse
from baconian.algo.misc.placeholder_input import PlaceholderInput
import baconian.algo.distribution.mvn as mvn

"""
logvar and logvar_speed is referred from https://github.com/pat-coady/trpo

"""


class NormalDistributionMLPPolicy(StochasticPolicy, PlaceholderInput):

    def __init__(self, env_spec: EnvSpec,
                 name: str,
                 name_scope: str,
                 mlp_config: list,
                 input_norm: np.ndarray = None,
                 output_norm: np.ndarray = None,
                 output_low: np.ndarray = None,
                 output_high: np.ndarray = None,
                 reuse=False,
                 distribution_tensors_tuple: tuple = None
                 ):
        StochasticPolicy.__init__(self, env_spec=env_spec, name=name, parameters=None)
        obs_dim = env_spec.flat_obs_dim
        action_dim = env_spec.flat_action_dim
        assert action_dim == mlp_config[-1]['N_UNITS']
        self.mlp_config = mlp_config
        self.input_norm = input_norm
        self.output_norm = output_norm
        self.output_low = output_low
        self.output_high = output_high
        self.mlp_config = mlp_config
        self.name_scope = name_scope

        mlp_kwargs = dict(
            reuse=reuse,
            input_norm=input_norm,
            output_norm=output_norm,
            output_low=output_low,
            output_high=output_high,
            mlp_config=mlp_config,
            name_scope=name_scope
        )
        ph_inputs = []
        if distribution_tensors_tuple is not None:
            self.mean_output = distribution_tensors_tuple[0][0]
            self.logvar_output = distribution_tensors_tuple[1][0]
            assert list(self.mean_output.shape)[-1] == action_dim
            assert list(self.logvar_output.shape)[-1] == action_dim
            self.mlp_net = None
        else:
            with tf.variable_scope(name_scope):
                self.state_input = tf.placeholder(shape=[None, obs_dim], dtype=tf.float32, name='state_ph')
                ph_inputs.append(self.state_input)
            self.mlp_net = MLP(input_ph=self.state_input,
                               net_name='normal_distribution_mlp_policy',
                               **mlp_kwargs)
            self.mean_output = self.mlp_net.output
            with tf.variable_scope(name_scope, reuse=reuse):
                with tf.variable_scope('norm_dist', reuse=reuse):
                    logvar_speed = (10 * self.mlp_config[-2]['N_UNITS']) // 48
                    logvar_output = tf.get_variable(name='normal_distribution_variance',
                                                    shape=[logvar_speed, self.mlp_config[-1]['N_UNITS']],
                                                    dtype=tf.float32)
                    self.logvar_output = tf.reduce_sum(logvar_output, axis=0)
        with tf.variable_scope(name_scope, reuse=reuse):
            self.action_input = tf.placeholder(shape=[None, action_dim], dtype=tf.float32, name='action_ph')
            ph_inputs.append(self.action_input)
            with tf.variable_scope('norm_dist', reuse=reuse):
                self.stddev_output = tf.exp(self.logvar_output / 2.0, name='std_dev')
                self.var_output = tf.exp(self.logvar_output, name='variance')
                self.action_output = mvn.sample(mean_p=self.mean_output,
                                                var_p=self.var_output,
                                                dims=self.action_space.flat_dim)
        var_list = get_tf_collection_var_list(scope='{}/norm_dist'.format(name_scope))
        if self.mlp_net:
            var_list += self.mlp_net.var_list

        self.parameters = ParametersWithTensorflowVariable(tf_var_list=sorted(list(set(var_list)),
                                                                              key=lambda x: x.name),
                                                           rest_parameters=dict(
                                                               state_input=self.state_input,
                                                               action_input=self.action_input,
                                                               **mlp_kwargs
                                                           ),
                                                           name='normal_distribution_mlp_tf_param')
        PlaceholderInput.__init__(self, parameters=self.parameters)
        self._log_prob = None
        self._prob = None
        self._entropy = None

    @overrides.overrides
    def forward(self, obs: (np.ndarray, list), sess=None, feed_dict=None, **kwargs):
        obs = make_batch(obs, original_shape=self.env_spec.obs_shape)
        feed_dict = feed_dict if feed_dict is not None else dict()
        feed_dict = {
            **feed_dict,
            self.state_input: obs,
            **self.parameters.return_tf_parameter_feed_dict()
        }
        sess = sess if sess else tf.get_default_session()
        res = sess.run(self.action_output, feed_dict=feed_dict)
        res = np.clip(res, a_min=self.env_spec.action_space.low, a_max=self.env_spec.action_space.high)
        return res

    @overrides.overrides
    def copy_from(self, obj) -> bool:
        return PlaceholderInput.copy_from(self, obj)

    def make_copy(self, **kwargs):
        kwargs = _get_copy_arg_with_tf_reuse(obj=self, kwargs=kwargs)
        copy_mlp_policy = NormalDistributionMLPPolicy(env_spec=self.env_spec,
                                                      input_norm=self.input_norm,
                                                      output_norm=self.output_norm,
                                                      output_low=self.output_low,
                                                      output_high=self.output_high,
                                                      mlp_config=self.mlp_config,
                                                      **kwargs)
        return copy_mlp_policy

    def compute_dist_info(self, name, sess=None, **kwargs) -> np.ndarray:
        # todo to be optimized, every time call the function will add computing tensor into the graph
        dist_info_map = {
            'log_prob': self.log_prob,
            'prob': self.prob,
            'entropy': self.entropy,
            'kl': self.kl,
        }
        if name not in dist_info_map.keys():
            raise ValueError("only support compute {} info".format(list(dist_info_map.keys())))
        sess = sess if sess else tf.get_default_session()
        if name in ['log_prob', 'prob']:
            if 'value' not in kwargs:
                raise ValueError('To compute {}, pass the parameter with key {} and type {} in'.format(name, 'value',
                                                                                                       np.ndarray.__name__))
            assert isinstance(kwargs['value'], np.ndarray)
        if name == 'kl':
            if 'other' not in kwargs:
                raise ValueError('To compute {}, pass the parameter with key {} and type {} in'.format(name, 'other',
                                                                                                       type(
                                                                                                           self).__name__))
            assert isinstance(kwargs['other'], type(self))
        if 'feed_dict' in kwargs:
            feed_dict = kwargs['feed_dict'] if 'feed_dict' in kwargs else None
            kwargs.pop('feed_dict')
        else:
            feed_dict = None

        return sess.run(dist_info_map[name](**kwargs), feed_dict=feed_dict)

    def kl(self, other, *args, **kwargs) -> tf.Tensor:
        if not isinstance(other, type(self)):
            raise TypeError()
        return mvn.kl(mean_p=self.mean_output, var_p=self.var_output,
                      mean_q=other.mean_output, var_q=other.var_output, dims=self.action_space.flat_dim)

    def log_prob(self, *args, **kwargs) -> tf.Tensor:
        if self._log_prob is None:
            self._log_prob = mvn.log_prob(variable_ph=self.action_input, mean_p=self.mean_output, var_p=self.var_output)
        return self._log_prob

    def prob(self, *args, **kwargs) -> tf.Tensor:
        if self._prob is None:
            self._prob = mvn.prob(variable_ph=self.action_input, mean_p=self.mean_output, var_p=self.var_output)
        return self._prob

    def entropy(self, *args, **kwargs) -> tf.Tensor:
        if self._entropy is None:
            self._entropy = mvn.entropy(self.mean_output, self.var_output, dims=self.action_space.flat_dim)
        return self._entropy

    def get_dist_info(self) -> tuple:
        res = (
            dict(shape=tuple(self.mean_output.shape.as_list()),
                 name='mean_output',
                 obj=self.mean_output,
                 dtype=self.mean_output.dtype),
            dict(shape=tuple(self.logvar_output.shape.as_list()),
                 name='logvar_output',
                 obj=self.logvar_output,
                 dtype=self.logvar_output.dtype),
        )
        for re in res:
            attr = getattr(self, re['name'])
            if id(attr) != id(re['obj']):
                raise ValueError('key name {} should be same as the obj {} name'.format(re['name'], re['obj']))
        return res

    def save(self, *args, **kwargs):
        return PlaceholderInput.save(self, *args, **kwargs)

    def load(self, *args, **kwargs):
        return PlaceholderInput.load(self, *args, **kwargs)


if __name__ == '__main__':
    from baconian.test.tests.test_rl.test_policy.test_mlp_norm_policy import TestNormalDistMLPPolicy
    import unittest

    unittest.TestLoader().loadTestsFromTestCase(TestNormalDistMLPPolicy)
    unittest.main()



"""
Reward functions in most of some model-based methods are tricky problems,
since model itself require a reward function from outside.

Most of the codebase tune the definition of the reward function differed from original one
without explicit notifying which bring the difficult for users to tune the algorithms without the effect
brought by reward functions.

In Baconian, we try to clarify this part, make sure the user is well aware of the effect of such implementations.

This is a work in progress.

"""

from baconian.algo.dynamics.reward_func.reward_func import RewardFunc
import numpy as np


class PendulumRewardFunc(RewardFunc):

    def __init__(self, name='pendulum_reward_func'):
        super().__init__(name)
        self.max_speed = 8
        self.max_torque = 2.
        self.dt = .05

    def __call__(self, state, action, new_state, **kwargs) -> float:
        th = state[0]
        thdot = state[1]
        u = np.clip(action, -self.max_torque, self.max_torque)[0]
        costs = angle_normalize(th) ** 2 + .1 * thdot ** 2 + .001 * (u ** 2)
        return float(-costs)

    def init(self):
        super().init()


def angle_normalize(x):
    return ((x + np.pi) % (2 * np.pi)) - np.pi


REWARD_FUNC_DICT = {
    'Pendulum-v0': PendulumRewardFunc
}



from baconian.core.core import Basic, EnvSpec, Env
from baconian.core.status import StatusWithSubInfo
import abc
from typeguard import typechecked
from baconian.common.logging import Recorder
from baconian.core.parameters import Parameters
from baconian.common.sampler.sample_data import TrajectoryData


class Algo(Basic):
    """
    Abstract class for algorithms
    """
    STATUS_LIST = ['CREATED', 'INITED', 'TRAIN', 'TEST']
    INIT_STATUS = 'CREATED'

    @typechecked
    def __init__(self, env_spec: EnvSpec, name: str = 'algo', warm_up_trajectories_number=0):
        """
        Constructor

        :param env_spec: environment specifications
        :type env_spec: EnvSpec
        :param name: name of the algorithm
        :type name: str
        :param warm_up_trajectories_number: how many trajectories used to warm up the training
        :type warm_up_trajectories_number: int
        """

        super().__init__(status=StatusWithSubInfo(obj=self), name=name)
        self.env_spec = env_spec
        self.parameters = Parameters(dict())
        self.recorder = Recorder(default_obj=self)
        self.warm_up_trajectories_number = warm_up_trajectories_number

    def init(self):
        """
        Initialization method, such as network random initialization in Tensorflow

        :return:
        """
        self._status.set_status('INITED')

    def warm_up(self, trajectory_data: TrajectoryData):
        """
        Use some data to warm up the algorithm, e.g., compute the mean/std-dev of the state to perform normalization.
        Data used in warm up process will not be added into the memory
        :param trajectory_data: TrajectoryData object
        :type trajectory_data: TrajectoryData

        :return: None
        """
        pass

    def train(self, *arg, **kwargs) -> dict:
        """
        Training API, specific arguments should be defined by each algorithms itself.

        :return: training results, e.g., loss
        :rtype: dict
        """

        self._status.set_status('TRAIN')
        return dict()

    def test(self, *arg, **kwargs) -> dict:
        """
        Testing API, most of the evaluation can be done by agent instead of algorithms, so this API can be skipped

        :return: test results, e.g., rewards
        :rtype: dict
        """

        self._status.set_status('TEST')
        return dict()

    @abc.abstractmethod
    def predict(self, *arg, **kwargs):
        """
        Predict function, given the obs as input, return the action, obs will be read as the first argument passed into
        this API, like algo.predict(obs=x, ...)

        :return: predicted action
        :rtype: np.ndarray
        """
        raise NotImplementedError

    @abc.abstractmethod
    def append_to_memory(self, *args, **kwargs):
        """
        For off-policy algorithm, use this API to append the data into replay buffer. samples will be read as the first
        argument passed into this API, like algo.append_to_memory(samples=x, ...)

        """
        raise NotImplementedError

    @property
    def is_training(self):
        """
        A boolean indicate the if the algorithm is in training status

        :return: True if in training
        :rtype: bool
        """
        return self.get_status()['status'] == 'TRAIN'

    @property
    def is_testing(self):
        """
        A boolean indicate the if the algorithm is in training status

        :return: True if in testing
        :rtype: bool
        """
        return self.get_status()['status'] == 'TEST'



import json_tricks as json
import os
import shutil
from baconian.common.error import *


def create_path(path, del_if_existed=True):
    if os.path.exists(path) is True and del_if_existed is False:
        raise FileExistsError()
    else:
        try:
            shutil.rmtree(path)
        except FileNotFoundError:
            pass
    os.makedirs(path)


def load_json(file_path) -> (dict, list):
    with open(file_path, 'r') as f:
        res = json.load(f)
        return res


def check_dir(path):
    if os.path.isdir(path) is False:
        raise LogPathOrFileNotExistedError('{} not existed'.format(path))


def check_file(path):
    if os.path.isfile(path) is False:
        raise LogPathOrFileNotExistedError('{} not existed'.format(path))


def save_to_json(obj: (list, dict), path=None, fp=None, file_name=None):
    jsonable_dict = convert_to_jsonable(dict_or_list=obj)
    if fp:
        json.dump(obj=jsonable_dict, fp=fp, indent=4, sort_keys=True)
        fp.close()
    else:
        if file_name is not None:
            path = os.path.join(path, file_name)
        with open(path, 'w') as f:
            json.dump(obj=obj, fp=f, indent=4, sort_keys=True)


def convert_to_jsonable(dict_or_list) -> (list, dict):
    if isinstance(dict_or_list, list):
        jsonable_dict = []
        for val in dict_or_list:
            if isinstance(val, (dict, list)):
                res = convert_to_jsonable(dict_or_list=val)
                jsonable_dict.append(res)
            else:
                f = open(os.devnull, 'w')
                try:
                    json.dump([val], f)
                except Exception:
                    jsonable_dict.append(str(val))
                else:
                    jsonable_dict.append(val)
                finally:
                    f.close()
        return jsonable_dict

    elif isinstance(dict_or_list, dict):
        jsonable_dict = dict()
        for key, val in dict_or_list.items():
            if isinstance(val, (dict, list)):
                res = convert_to_jsonable(dict_or_list=val)
                jsonable_dict[key] = res
            else:
                f = open(os.devnull, 'w')
                try:
                    json.dump([val], f)
                except Exception:
                    jsonable_dict[key] = str(val)
                else:
                    jsonable_dict[key] = val
                finally:
                    f.close()
        return jsonable_dict


def convert_dict_to_csv(log_dict):
    """
    This function will convert a log dict into csv file by recursively find the list in the
    dict and save the list with its key as a single csv file, the
    :param log_dict:
    :return: list, as the dict as each element with
    {'csv_file_name', 'csv_keys', 'csv_row_data'}
    """
    raise NotImplementedError



from baconian.common.sampler.sample_data import TransitionData
from baconian.test.tests.set_up.setup import TestWithAll
import numpy as np


class TestDynamicsModel(TestWithAll):

    def test_mlp_dynamics_model(self):
        mlp_dyna, local = self.create_continue_dynamics_model(name='mlp_dyna_model')
        env = local['env']
        env_spec = local['env_spec']
        env.reset()
        mlp_dyna.init()
        for i in range(100):
            mlp_dyna.step(action=np.array(env_spec.action_space.sample()),
                          state=env_spec.obs_space.sample())
        data = TransitionData(env_spec)
        st = env.get_state()
        for i in range(10):
            ac = env_spec.action_space.sample()
            new_st, re, done, info = env.step(action=ac)
            data.append(state=st,
                        action=ac,
                        new_state=new_st,
                        done=done,
                        reward=re)
            st = new_st
        print(mlp_dyna.train(batch_data=data, train_iter=10))
        mlp_dyna_2, _ = self.create_continue_dynamics_model(name='model_2')
        mlp_dyna_2.init()
        self.assert_var_list_at_least_not_equal(var_list1=mlp_dyna.parameters('tf_var_list'),
                                                var_list2=mlp_dyna_2.parameters('tf_var_list'))

        self.assert_var_list_id_no_equal(var_list1=mlp_dyna.parameters('tf_var_list'),
                                         var_list2=mlp_dyna_2.parameters('tf_var_list'))

        mlp_dyna_2.init(source_obj=mlp_dyna)

        self.assert_var_list_equal(var_list1=mlp_dyna.parameters('tf_var_list'),
                                   var_list2=mlp_dyna_2.parameters('tf_var_list'))

        self.assert_var_list_id_no_equal(var_list1=mlp_dyna.parameters('tf_var_list'),
                                         var_list2=mlp_dyna_2.parameters('tf_var_list'))

        mlp_dyna_2.copy_from(mlp_dyna)

        self.assert_var_list_equal(var_list1=mlp_dyna.parameters('tf_var_list'),
                                   var_list2=mlp_dyna_2.parameters('tf_var_list'))

        self.assert_var_list_id_no_equal(var_list1=mlp_dyna.parameters('tf_var_list'),
                                         var_list2=mlp_dyna_2.parameters('tf_var_list'))



# Date: 3/30/19
# Author: Luke
# Project: baconian-internal

"""
A simple example to show how to build up an experiment with ppo training and testing on Pendulum-v0
"""
from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.algo.value_func import MLPVValueFunc
from baconian.algo.ppo import PPO
from baconian.algo.policy.normal_distribution_mlp import NormalDistributionMLPPolicy
from baconian.core.agent import Agent
from baconian.algo.misc import EpsilonGreedy
from baconian.core.experiment import Experiment
from baconian.core.flow.train_test_flow import create_train_test_flow
from baconian.config.global_config import GlobalConfig
from baconian.core.status import get_global_status_collect
from baconian.common.schedules import PeriodicalEventSchedule


def task_fn():
    env = make('Pendulum-v0')
    name = 'demo_exp_'
    env_spec = EnvSpec(obs_space=env.observation_space,
                       action_space=env.action_space)

    mlp_v = MLPVValueFunc(env_spec=env_spec,
                          name_scope=name + 'mlp_v',
                          name=name + 'mlp_v',
                          mlp_config=[
                              {
                                  "ACT": "RELU",
                                  "B_INIT_VALUE": 0.0,
                                  "NAME": "1",
                                  "N_UNITS": 16,
                                  "L1_NORM": 0.01,
                                  "L2_NORM": 0.01,
                                  "TYPE": "DENSE",
                                  "W_NORMAL_STDDEV": 0.03
                              },
                              {
                                  "ACT": "LINEAR",
                                  "B_INIT_VALUE": 0.0,
                                  "NAME": "OUPTUT",
                                  "N_UNITS": 1,
                                  "TYPE": "DENSE",
                                  "W_NORMAL_STDDEV": 0.03
                              }
                          ])

    policy = NormalDistributionMLPPolicy(env_spec=env_spec,
                                         name_scope=name + 'mlp_policy',
                                         name=name + 'mlp_policy',
                                         mlp_config=[
                                             {
                                                 "ACT": "RELU",
                                                 "B_INIT_VALUE": 0.0,
                                                 "NAME": "1",
                                                 "L1_NORM": 0.01,
                                                 "L2_NORM": 0.01,
                                                 "N_UNITS": 16,
                                                 "TYPE": "DENSE",
                                                 "W_NORMAL_STDDEV": 0.03
                                             },
                                             {
                                                 "ACT": "LINEAR",
                                                 "B_INIT_VALUE": 0.0,
                                                 "NAME": "OUPTUT",
                                                 "N_UNITS": env_spec.flat_action_dim,
                                                 "TYPE": "DENSE",
                                                 "W_NORMAL_STDDEV": 0.03
                                             }
                                         ],
                                         reuse=False)

    ppo = PPO(
        env_spec=env_spec,
        config_or_config_dict={
            "gamma": 0.995,
            "lam": 0.98,
            "policy_train_iter": 10,
            "value_func_train_iter": 10,
            "clipping_range": None,
            "beta": 1.0,
            "eta": 50,
            "log_var_init": -1.0,
            "kl_target": 0.003,
            "policy_lr": 0.01,
            "value_func_lr": 0.01,
            "value_func_train_batch_size": 10,
            "lr_multiplier": 1.0
        },
        value_func=mlp_v,
        stochastic_policy=policy,
        name=name + 'ppo'
    )
    agent = Agent(env=env, env_spec=env_spec,
                  algo=ppo,
                  algo_saving_scheduler=PeriodicalEventSchedule(
                      t_fn=lambda: get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
                      trigger_every_step=20,
                      after_t=10),
                  name=name + 'agent',
                  exploration_strategy=EpsilonGreedy(action_space=env_spec.action_space,
                                                     init_random_prob=0.5))
    flow = create_train_test_flow(
        test_every_sample_count=10,
        train_every_sample_count=10,
        start_test_after_sample_count=5,
        start_train_after_sample_count=5,
        train_func_and_args=(agent.train, (), dict()),
        test_func_and_args=(agent.test, (), dict(sample_count=10)),
        sample_func_and_args=(agent.sample, (), dict(sample_count=100,
                                                     env=agent.env,
                                                     sample_type='trajectory',
                                                     store_flag=True))
    )

    experiment = Experiment(
        tuner=None,
        env=env,
        agent=agent,
        flow=flow,
        name=name
    )
    experiment.run()


from baconian.core.experiment_runner import single_exp_runner

GlobalConfig().set('DEFAULT_LOG_PATH', './log_path')
single_exp_runner(task_fn, del_if_log_path_existed=True)



from baconian.envs.gym_env import make
from baconian.core.core import EnvSpec
from baconian.test.tests.set_up.setup import BaseTestCase
from baconian.common.data_pre_processing import *
import numpy as np


class TestDataPreProcessing(BaseTestCase):
    def test_min_max(self):
        for env in (make('Pendulum-v0'), make('Acrobot-v1'), make('RoboschoolAnt-v1')):
            for sample_space in (env.observation_space, env.action_space):
                sample_fn = sample_space.sample
                dims = sample_space.flat_dim
                try:
                    print("test {} with sample {} dims {}".format(env, sample_fn, dims))
                    # test batch scaler
                    min_max = BatchMinMaxScaler(dims=dims)
                    data_list = []
                    for i in range(100):
                        data_list.append(sample_fn())
                    data = min_max.process(np.array(data_list))
                    self.assertTrue(np.greater_equal(np.ones(dims),
                                                     data).all())
                    self.assertTrue(np.less_equal(np.zeros(dims),
                                                  data).all())
                    # test batch scaler with given range
                    min_max = BatchMinMaxScaler(dims=dims,
                                                desired_range=(np.ones(dims) * -1.0,
                                                               np.ones(dims) * 5.0))
                    data_list = []
                    for i in range(100):
                        data_list.append(sample_fn())
                    data = min_max.process(np.array(data_list))
                    self.assertTrue(np.greater_equal(np.ones(dims) * 5.0,
                                                     data).all())
                    self.assertTrue(np.less_equal(np.ones(dims) * -1.0,
                                                  data).all())
                    self.assertEqual(np.max(data), 5.0)
                    self.assertEqual(np.min(data), -1.0)
                    data = min_max.inverse_process(data)
                    self.assertTrue(np.isclose(data, np.array(data_list)).all())

                    # test batch scaler with given range and given initial data
                    data_list = []
                    for i in range(100):
                        data_list.append(sample_fn())

                    min_max = RunningMinMaxScaler(dims=dims,
                                                  desired_range=(np.ones(dims) * -1.0,
                                                                 np.ones(dims) * 5.0),
                                                  init_data=np.array(data_list))

                    data = min_max.process(np.array(data_list))
                    self.assertTrue(np.greater_equal(np.ones(dims) * 5.0,
                                                     data).all())
                    self.assertTrue(np.less_equal(np.ones(dims) * -1.0,
                                                  data).all())
                    self.assertEqual(np.max(data), 5.0)
                    self.assertEqual(np.min(data), -1.0)

                    # test batch scaler with given range and given initial min and max
                    data_list = []
                    for i in range(100):
                        data_list.append(sample_fn())

                    min_max = RunningMinMaxScaler(dims=dims,
                                                  desired_range=(np.ones(dims) * -1.0,
                                                                 np.ones(dims) * 5.0),
                                                  init_min=np.min(np.array(data_list), axis=0),
                                                  init_max=np.max(np.array(data_list), axis=0))

                    data = min_max.process(np.array(data_list))
                    self.assertTrue(np.greater_equal(np.ones(dims) * 5.0,
                                                     data).all())
                    self.assertTrue(np.less_equal(np.ones(dims) * -1.0,
                                                  data).all())
                    self.assertEqual(np.max(data), 5.0)
                    self.assertEqual(np.min(data), -1.0)

                    # test update function by a larger range of data
                    pre_min = np.min(np.array(data_list), axis=0)
                    pre_max = np.max(np.array(data_list), axis=0)
                    data_list = np.array(data_list) * 2.0
                    min_max.update_scaler(data_list)
                    self.assertTrue(np.equal(pre_min * 2.0, min_max._min).all())
                    self.assertTrue(np.equal(pre_max * 2.0, min_max._max).all())
                except ShapeNotCompatibleError as e:
                    from baconian.common.spaces import Box
                    if isinstance(sample_space, Box):
                        raise ValueError
                    else:
                        pass

    def test_standard_scaler(self):
        for env in (make('Pendulum-v0'), make('Acrobot-v1'), make('RoboschoolAnt-v1')):
            for sample_space in (env.observation_space, env.action_space):
                sample_fn = sample_space.sample
                dims = sample_space.flat_dim
                try:
                    # test batch standard scaler
                    standard_scaler = BatchStandardScaler(dims=dims)
                    data_list = []
                    for i in range(100):
                        data_list.append(sample_fn())
                    data = standard_scaler.process(np.array(data_list))
                    self.assertTrue(np.isclose(np.mean(data, axis=0), 0.0).all())
                    # TODO a theoretical bound should be given
                    # self.assertTrue(np.isclose(np.var(data, axis=0), 1.0, atol=0.04).all())
                    data = standard_scaler.inverse_process(data)
                    self.assertTrue(np.isclose(data, np.array(data_list)).all())

                    # test running standard scaler
                    standard_scaler = RunningStandardScaler(dims=dims)
                    data_list = []
                    for i in range(100):
                        data_list.append(sample_fn())
                    standard_scaler.update_scaler(np.array(data_list))
                    self.assertEqual(standard_scaler._data_count, 100)
                    data = standard_scaler.process(np.array(data_list))
                    self.assertTrue(np.isclose(np.mean(data, axis=0), 0.0).all())

                    # TODO a theoretical bound should be given
                    # self.assertTrue(np.isclose(np.var(data, axis=0), 1.0, atol=0.04).all())
                    # test update function
                    new_data_list = []
                    for i in range(100):
                        new_data_list.append(sample_fn())
                    standard_scaler.update_scaler(np.array(new_data_list))
                    self.assertEqual(standard_scaler._data_count, 200)

                    data_list += new_data_list
                    data = standard_scaler.process(np.array(data_list))
                    self.assertTrue(np.isclose(np.mean(data, axis=0), 0.0).all())

                    # TODO a theoretical bound should be given
                    # self.assertTrue(np.isclose(np.var(data, axis=0), 1.0, atol=0.04).all())

                    # test running scaler with given data
                    data_list = []
                    for i in range(100):
                        data_list.append(sample_fn())
                    standard_scaler = RunningStandardScaler(dims=dims,
                                                            init_data=np.array(data_list))

                    self.assertEqual(standard_scaler._data_count, 100)
                    data = standard_scaler.process(np.array(data_list))
                    self.assertTrue(np.isclose(np.mean(data, axis=0), 0.0).all())
                    # TODO a theoretical bound should be given
                    # self.assertTrue(np.isclose(np.var(data, axis=0), 1.0, atol=0.04).all())
                    # test update of running scaler with given data
                    new_data_list = []
                    for i in range(100):
                        new_data_list.append(sample_fn())
                    standard_scaler.update_scaler(np.array(new_data_list))
                    self.assertEqual(standard_scaler._data_count, 200)

                    data_list += new_data_list
                    data = standard_scaler.process(np.array(data_list))
                    self.assertTrue(np.isclose(np.mean(data, axis=0), 0.0).all())

                    # TODO a theoretical bound should be given
                    # self.assertTrue(np.isclose(np.var(data, axis=0), 1.0, atol=0.04).all())

                    # test running scaler with given initial mean, var.
                    data_list = []
                    for i in range(100):
                        data_list.append(sample_fn())
                    standard_scaler = RunningStandardScaler(dims=dims,
                                                            init_mean=np.mean(data_list, axis=0),
                                                            init_var=np.var(data_list, axis=0),
                                                            init_mean_var_data_count=100)

                    self.assertEqual(standard_scaler._data_count, 100)
                    data = standard_scaler.process(np.array(data_list))
                    self.assertTrue(np.isclose(np.mean(data, axis=0), 0.0).all())
                    # TODO a theoretical bound should be given
                    # self.assertTrue(np.isclose(np.var(data, axis=0), 1.0, atol=0.04).all())

                    new_data_list = []
                    for i in range(100):
                        new_data_list.append(sample_fn())
                    standard_scaler.update_scaler(np.array(new_data_list))
                    self.assertEqual(standard_scaler._data_count, 200)

                    data_list += new_data_list
                    data = standard_scaler.process(np.array(data_list))
                    self.assertTrue(np.isclose(np.mean(data, axis=0), 0.0).all())

                    # TODO a theoretical bound should be given
                    # self.assertTrue(np.isclose(np.var(data, axis=0), 1.0, atol=0.04).all())
                except ShapeNotCompatibleError as e:
                    from baconian.common.spaces import Box
                    if isinstance(sample_space, Box):
                        raise ValueError
                    else:
                        pass



from baconian.envs.gym_env import GymEnv
from baconian.test.tests.set_up.setup import TestWithLogSet
from gym import make
import os
import platform
from pathlib import Path

_PLATFORM = platform.system()
try:
    _PLATFORM_SUFFIX = {
        "Linux": "linux",
        "Darwin": "macos",
        "Windows": "win64"
    }[_PLATFORM]
except KeyError:
    raise OSError("Unsupported platform: {}".format(_PLATFORM))

have_mujoco_flag = True
try:
    import mujoco_py
except Exception:
    have_mujoco_flag = False

if have_mujoco_flag:
    os.environ['LD_LIBRARY_PATH'] = os.environ.get('LD_LIBRARY_PATH', '') \
                                    + ':' + str(Path.home()) + '/.mujoco/mujoco200_{}/bin'.format(_PLATFORM_SUFFIX)

    os.environ['MUJOCO_PY_MUJOCO_PATH'] = os.environ.get('MUJOCO_PY_MUJOCO_PATH', '') \
                                          + str(Path.home()) + '/.mujoco/mujoco200_{}'.format(_PLATFORM_SUFFIX)


class TestEnv(TestWithLogSet):
    def test_gym_env(self):
        if have_mujoco_flag:
            a = GymEnv('FetchPickAndPlace-v1')
            a.set_status('TRAIN')
            self.assertEqual(a.total_step_count_fn(), 0)
            self.assertEqual(a._last_reset_point, 0)
            a.init()
            a.seed(10)
            a.reset()
            self.assertEqual(a.total_step_count_fn(), 0)
            self.assertEqual(a._last_reset_point, 0)
            for i in range(1000):
                new_st, re, done, _ = a.step(action=a.action_space.sample())
                self.assertEqual(a.total_step_count_fn(), i + 1)
                if done is True:
                    a.reset()
                    self.assertEqual(a._last_reset_point, a.total_step_count_fn())
                    self.assertEqual(a._last_reset_point, i + 1)



import typeguard as tg
from baconian.core.core import EnvSpec
import overrides
import tensorflow as tf
from baconian.tf.tf_parameters import ParametersWithTensorflowVariable
from baconian.tf.mlp import MLP
from baconian.common.special import *
from baconian.algo.utils import _get_copy_arg_with_tf_reuse
from baconian.algo.misc.placeholder_input import PlaceholderInput
from baconian.algo.value_func import VValueFunction


class MLPVValueFunc(VValueFunction, PlaceholderInput):
    """
    Multi Layer Q Value Function, based on Tensorflow, take the state and action as input,
    return the Q value for all action/ input action.
    """

    @tg.typechecked
    def __init__(self,
                 env_spec: EnvSpec,
                 name_scope: str,
                 name: str,
                 mlp_config: list,
                 state_input: tf.Tensor = None,
                 reuse=False,
                 input_norm: np.ndarray = None,
                 output_norm: np.ndarray = None,
                 output_low: np.ndarray = None,
                 output_high: np.ndarray = None,
                 ):
        with tf.variable_scope(name_scope):
            state_input = state_input if state_input is not None else tf.placeholder(
                shape=[None, env_spec.flat_obs_dim],
                dtype=tf.float32,
                name='state_ph')

        mlp_input_ph = state_input
        mlp_kwargs = dict(
            reuse=reuse,
            mlp_config=mlp_config,
            input_norm=input_norm,
            output_norm=output_norm,
            output_high=output_high,
            output_low=output_low,
            name_scope=name_scope
        )
        mlp_net = MLP(input_ph=mlp_input_ph,
                      net_name='mlp',
                      **mlp_kwargs)
        parameters = ParametersWithTensorflowVariable(tf_var_list=mlp_net.var_list,
                                                      rest_parameters=mlp_kwargs,
                                                      name='mlp_v_value_function_tf_param')
        VValueFunction.__init__(self,
                                env_spec=env_spec,
                                state_input=state_input,
                                name=name,
                                parameters=None)
        PlaceholderInput.__init__(self, parameters=parameters)

        self.name_scope = name_scope
        self.mlp_config = mlp_config
        self.input_norm = input_norm
        self.output_norm = output_norm
        self.output_low = output_low
        self.output_high = output_high
        self.state_input = state_input
        self.mlp_input_ph = mlp_input_ph
        self.mlp_net = mlp_net
        self.v_tensor = self.mlp_net.output

    @overrides.overrides
    def copy_from(self, obj: PlaceholderInput) -> bool:
        return PlaceholderInput.copy_from(self, obj)

    @overrides.overrides
    def forward(self, obs: (np.ndarray, list), sess=None,
                feed_dict=None, *args,
                **kwargs):
        obs = make_batch(obs, original_shape=self.env_spec.obs_shape)
        feed_dict = feed_dict if feed_dict is not None else dict()
        sess = sess if sess else tf.get_default_session()
        feed_dict = {
            self.state_input: obs,
            **feed_dict,
            **self.parameters.return_tf_parameter_feed_dict()
        }
        v = sess.run(self.v_tensor,
                     feed_dict=feed_dict)
        return v

    def init(self, source_obj=None):
        self.parameters.init()
        if source_obj:
            self.copy_from(obj=source_obj)

    def make_copy(self, *args, **kwargs):
        kwargs = _get_copy_arg_with_tf_reuse(obj=self, kwargs=kwargs)

        copy_mlp_v_value = MLPVValueFunc(env_spec=self.env_spec,
                                         input_norm=self.input_norm,
                                         output_norm=self.output_norm,
                                         output_low=self.output_low,
                                         output_high=self.output_high,
                                         mlp_config=self.mlp_config,
                                         **kwargs)
        return copy_mlp_v_value

    def save(self, *args, **kwargs):
        return PlaceholderInput.save(self, *args, **kwargs)

    def load(self, *args, **kwargs):
        return PlaceholderInput.load(self, *args, **kwargs)



import os
import platform

from pathlib import Path

_PLATFORM = platform.system()
try:
    _PLATFORM_SUFFIX = {
        "Linux": "linux",
        "Darwin": "macos",
        "Windows": "win64"
    }[_PLATFORM]
except KeyError:
    raise OSError("Unsupported platform: {}".format(_PLATFORM))

# TODO potential bug here if mujoco py change default path in the future

os.environ['LD_LIBRARY_PATH'] = os.environ.get('LD_LIBRARY_PATH', '') \
                                + ':' + str(Path.home()) + '/.mujoco/mujoco200/bin'

if os.environ.get('MUJOCO_PY_MUJOCO_PATH', '') == ' ':
    os.environ['MUJOCO_PY_MUJOCO_PATH'] = str(Path.home()) + '/.mujoco/mujoco200'

if os.environ.get('MUJOCO_PY_MJKEY_PATH', '') == ' ':
    os.environ['MUJOCO_PY_MJKEY_PATH'] = str(Path.home()) + '/.mujoco/mujoco200'

os.environ['MJKEY_PATH'] = os.environ.get('MUJOCO_PY_MJKEY_PATH', '')
os.environ['MJLIB_PATH'] = os.environ.get('MUJOCO_PY_MUJOCO_PATH', '')

# TODO disable the rendering temporarily
os.environ['DISABLE_MUJOCO_RENDERING'] = '1'

have_mujoco_flag = True
try:
    from dm_control import mujoco
    from gym.envs.mujoco import mujoco_env
    from dm_control import suite

    from dm_control.rl.specs import ArraySpec
    from dm_control.rl.specs import BoundedArraySpec

    from collections import OrderedDict
except Exception:
    have_mujoco_flag = False



import glob
import os
import json_tricks as json
import numpy as np


def get_reward_json(root_dir_list, sub_dir, key, index_key):
    all_res = []
    for rt in root_dir_list:
        with open(file=os.path.join(rt, sub_dir)) as f:
            res = json.load(f)[key]
            val = [rr['log_val'] for rr in res]

            index = [rr[index_key] for rr in res]
            all_res.append((val, index))
            print(val[-5:], rt)
    aver = 0.0
    for re in all_res:
        aver += float(np.mean(re[0][-5:]))
    aver /= len(all_res)
    print(aver)


if __name__ == '__main__':
    get_reward_json(
        root_dir_list=glob.glob('/home/dls/CAP/baconian-internal/benchmark/benchmark_log/Pendulum-v0/dyna/**/*'),
        sub_dir='record/benchmark_agent/TEST/log.json',
        key='sum_reward',
        index_key='predict_counter'
    )



from baconian.envs.dmcontrol_env import have_mujoco_flag
import unittest


class TestEnv(unittest.TestCase):
    def test_dmcontrol_env(self):
        if have_mujoco_flag:
            from baconian.envs.dmcontrol_env import DMControlEnv
            a = DMControlEnv('cartpole', 'swingup')
            a.set_status('TRAIN')
            self.assertEqual(a.total_step_count_fn(), 0)
            self.assertEqual(a._last_reset_point, 0)
            a.init()
            a.seed(10)
            a.reset()
            self.assertEqual(a.total_step_count_fn(), 0)
            self.assertEqual(a._last_reset_point, 0)
            for i in range(1000):
                new_st, re, done, _ = a.step(action=a.action_space.sample())
                self.assertEqual(a.total_step_count_fn(), i + 1)
                if done is True:
                    a.reset()
                    self.assertEqual(a._last_reset_point, a.total_step_count_fn())
                    self.assertEqual(a._last_reset_point, i + 1)



import gym
import numpy as np
from typeguard import typechecked

from baconian.common.spaces import Space
from baconian.common.special import flat_dim, flatten

from baconian.common.logging import Recorder
from baconian.config.global_config import GlobalConfig
from baconian.core.status import *
from baconian.core.util import register_name_globally, init_func_arg_record_decorator

"""
This module contains the some core classes of baconian
"""


class Basic(object):
    """ Basic class within the whole framework"""
    STATUS_LIST = GlobalConfig().DEFAULT_BASIC_STATUS_LIST
    INIT_STATUS = GlobalConfig().DEFAULT_BASIC_INIT_STATUS
    required_key_dict = ()
    allow_duplicate_name = False

    def __init__(self, name: str, status=None):
        """
        Init a new Basic instance.

        :param name: name of the object, can be determined to generate log path, handle tensorflow name scope etc.
        :type name: str
        :param status: A status instance :py:class:`~baconian.core.status.Status` to indicate the status of the object
        :type status: Status
        """

        if not status:
            self._status = Status(self)
        else:
            self._status = status
        self._name = name
        register_name_globally(name=name, obj=self)

    def init(self, *args, **kwargs):
        """Initialize the object"""
        raise NotImplementedError

    def get_status(self) -> dict:
        """ Return the object's status, a dictionary."""
        return self._status.get_status()

    def set_status(self, val):
        """ Set the object's status."""
        self._status.set_status(val)

    @property
    def name(self):
        """ The name(id) of object, a string."""
        return self._name

    @property
    def status_list(self):
        """ Status list of the object, ('TRAIN', 'TEST')."""
        return self.STATUS_LIST

    def save(self, *args, **kwargs):
        """ Save the parameters in training checkpoints."""
        raise NotImplementedError

    def load(self, *args, **kwargs):
        """ Load the parameters from training checkpoints."""
        raise NotImplementedError


class Env(gym.Env, Basic):
    """
    Abstract class for environment
    """
    key_list = ()
    STATUS_LIST = ('JUST_RESET', 'INITED', 'TRAIN', 'TEST', 'CREATED')
    INIT_STATUS = 'CREATED'

    @typechecked
    def __init__(self, name: str = 'env', copy_from_env=None):
        super(Env, self).__init__(status=StatusWithSubInfo(obj=self), name=name)
        self.action_space = None
        self.observation_space = None
        self.trajectory_level_step_count = 0
        self.recorder = Recorder(default_obj=self)
        self._last_reset_point = 0
        self.total_step_count_fn = lambda: self._status.group_specific_info_key(info_key='step', group_way='sum')
        self.env_spec = None
        if copy_from_env:
            assert isinstance(copy_from_env, Env)
            self.action_space = copy_from_env.action_space
            self.observation_space = copy_from_env.observation_space
            self.trajectory_level_step_count = copy_from_env.trajectory_level_step_count
            self.trajectory_level_step_count = copy_from_env._last_reset_point
            self.env_spec = copy_from_env.env_spec

    @register_counter_info_to_status_decorator(increment=1, info_key='step', under_status=('TRAIN', 'TEST'),
                                               ignore_wrong_status=True)
    def step(self, action):
        """

        :param action: agent's action, the environment will react responding to action
        :type action: method
        """
        self.trajectory_level_step_count += 1
        pass

    @register_counter_info_to_status_decorator(increment=1, info_key='reset', under_status='JUST_RESET')
    def reset(self):
        """ Set the status to 'JUST_RESET', and update new reset point"""
        self._status.set_status('JUST_RESET')
        self._last_reset_point = self.total_step_count_fn()
        self.trajectory_level_step_count = 0

    @register_counter_info_to_status_decorator(increment=1, info_key='init', under_status='INITED')
    def init(self):
        """ Set the status to 'INITED'."""
        self._status.set_status('INITED')

    def get_state(self):
        """ Get the status of the environment."""
        raise NotImplementedError

    def seed(self, seed=None):
        """

        :param seed: seed to generate random number
        :type seed: int
        :return: seed of the unwrapped environment
        :rtype: int
        """
        return self.unwrapped.seed(seed=seed)


class EnvSpec(object):
    @init_func_arg_record_decorator()
    @typechecked
    def __init__(self, obs_space: Space, action_space: Space):
        self._obs_space = obs_space
        self._action_space = action_space

    @property
    def obs_shape(self):
        obs_shape = tuple(np.array(self.obs_space.sample()).shape)
        if len(obs_shape) == 0:
            obs_shape = (1,)
        return obs_shape

    @property
    def action_shape(self):
        action_shape = tuple(np.array(self.action_space.sample()).shape)
        if len(action_shape) == 0:
            action_shape = ()
        return action_shape

    @property
    def obs_space(self):
        """

        :return: Observation space of environment
        :rtype: Space
        """
        return self._obs_space

    @property
    def action_space(self):
        """

        :return: Action space of environment
        :rtype: Space
        """
        return self._action_space

    @obs_space.setter
    def obs_space(self, s: Space):
        self._obs_space = s

    @action_space.setter
    def action_space(self, s: Space):
        self._action_space = s

    @property
    def flat_obs_dim(self) -> int:
        """

        :return: the dimension(length) of flatten observation space
        :rtype: int
        """
        return int(flat_dim(self.obs_space))

    @property
    def flat_action_dim(self) -> int:
        """

        :return: the dimension(length) of flatten action space
        :rtype: int
        """
        return int(flat_dim(self.action_space))

    @staticmethod
    def flat(space: Space, obs_or_action: (np.ndarray, list)):
        """
        flat the input obs or action
        :param space: space of environment
        :type space: Space
        :param obs_or_action: action or observation space
        :type obs_or_action: (np.ndarray, list)
        :return: flatten action or observation space
        :rtype: Space
        """
        return flatten(space, obs_or_action)

    def flat_action(self, action: (np.ndarray, list)):
        """

        :param action: action taken by agent
        :type action: (np.ndarray, list)
        :return: flatten action parameter
        :rtype: np.ndarray
        """
        return flatten(self.action_space, action)

    def flat_obs(self, obs: (np.ndarray, list)):
        """

        :param obs: observation of the agent
        :type obs:  (np.ndarray, list)
        :return: flatten observation parameter
        :rtype: np.ndarray
        """
        return flatten(self.obs_space, obs)



from baconian.core.core import EnvSpec
from baconian.algo.dynamics.dynamics_model import TrainableDyanmicsModel, LocalDyanmicsModel
import gpflow
import numpy as np
from baconian.common.sampler.sample_data import TransitionData
from baconian.algo.dynamics.third_party.mgpr import MGPR
import tensorflow as tf
from baconian.tf.util import *
from baconian.tf.tf_parameters import ParametersWithTensorflowVariable
from baconian.common.data_pre_processing import DataScaler


class GaussianProcessDyanmicsModel(LocalDyanmicsModel, TrainableDyanmicsModel):
    kernel_type_dict = {
        'RBF': (gpflow.kernels.RBF, dict(ARD=True))
    }

    """
    Dynamics approximated by multivariate gaussian process model based GPflow package.
    Mostly refer the implementation of  PILCO repo in https://github.com/nrontsis/PILCO
    """

    def __init__(self, env_spec: EnvSpec, batch_data, init_state=None,
                 name_scope='gp_dynamics_model', name='gp_dynamics_model',
                 gp_kernel_type='RBF'):
        if gp_kernel_type not in self.kernel_type_dict.keys():
            raise TypeError(
                'Not supported {} kernel, choose from {}'.format(gp_kernel_type, list(self.kernel_type_dict.keys())))
        parameters = ParametersWithTensorflowVariable(tf_var_list=[],
                                                      rest_parameters=dict(),
                                                      name='{}_param'.format(name),
                                                      require_snapshot=False)
        super().__init__(env_spec=env_spec, parameters=parameters, init_state=init_state, name=name)
        self.name_scope = name_scope
        state_action_data = np.hstack((batch_data.state_set, batch_data.action_set))
        delta_state_data = batch_data.new_state_set - batch_data.state_set
        with tf.variable_scope(self.name_scope):
            self.mgpr_model = MGPR(name='mgpr', action_dim=env_spec.flat_action_dim,
                                   x=state_action_data, y=delta_state_data,
                                   state_dim=env_spec.flat_obs_dim)
        var_list = get_tf_collection_var_list(key=tf.GraphKeys.GLOBAL_VARIABLES,
                                              scope=self.name_scope)
        self.parameters.set_tf_var_list(tf_var_list=sorted(list(set(var_list)), key=lambda x: x.name))

    def init(self):
        super().init()

    def _state_transit(self, state, action, required_var=False, **kwargs):
        deltas, vars = self.mgpr_model.predict(x=np.expand_dims(np.concatenate([state, action], axis=0), axis=0))
        if required_var is True:
            return np.squeeze(deltas) + state, np.squeeze(vars)
        else:
            return np.squeeze(deltas) + state

    def copy_from(self, obj) -> bool:
        raise NotImplementedError

    def make_copy(self):
        raise NotImplementedError

    def train(self, restart=1, batch_data: TransitionData = None, *kwargs):
        if batch_data:
            state_action_data = np.hstack((batch_data.state_set, batch_data.action_set))
            delta_state_data = batch_data.new_state_set - batch_data.state_set
            self.mgpr_model.set_XY(X=state_action_data, Y=delta_state_data)
        self.mgpr_model.optimize(restarts=restart)



from typeguard import typechecked
from baconian.config.dict_config import DictConfig
import abc
from baconian.common.logging import Logger
import baconian.common.files as files
import os
from baconian.common.schedules import Scheduler
from copy import deepcopy, copy


class Parameter(object):
    # TODO
    def __init__(self):
        pass


class Parameters(object):
    """
    A class that handle all parameters of a certain rl, to be better support in the future version.
    Currently, just a very simple implementation
    """

    @typechecked
    def __init__(self, parameters: dict,
                 source_config: DictConfig = None,
                 name='parameters',
                 to_scheduler_param_tuple: tuple = None,
                 default_save_param_key=None):
        self._parameters = parameters
        self.name = name
        self._source_config = source_config if source_config else DictConfig(required_key_dict=dict(),
                                                                             config_dict=dict())
        self.default_save_param_key = default_save_param_key
        self._scheduler_info_dict = dict()
        self.to_scheduler_param_list = to_scheduler_param_tuple

    def __call__(self, key=None):
        if key:
            if key in self._scheduler_info_dict:
                new_val = self._scheduler_info_dict[key]['scheduler'].value()
                if key in self._parameters:
                    self._parameters[key] = new_val
                else:
                    self._source_config.set(key, new_val)
                return new_val
            if isinstance(self._parameters, dict):
                if key in self._parameters:
                    return self._parameters[key]
                else:
                    return self._source_config(key)
            else:
                raise ValueError('parameters is not dict')
        else:
            raise KeyError('specific a key to call {}'.format(type(self).__name__))

    def __getitem__(self, item):
        return self.__call__(key=item)

    def init(self):
        if self.to_scheduler_param_list:
            for val_dict in self.to_scheduler_param_list:
                self.set_scheduler(**val_dict)

    def copy_from(self, source_parameter):
        if not isinstance(source_parameter, type(self)):
            raise TypeError()

        self._update_dict(source_dict=source_parameter._parameters,
                          target_dict=self._parameters)
        self._source_config.config_dict = source_parameter._source_config.config_dict
        self.default_save_param_key = copy(source_parameter.default_save_param_key)
        if source_parameter.to_scheduler_param_list:
            self._scheduler_info_dict = dict()
            self.to_scheduler_param_list = copy(source_parameter.to_scheduler_param_list)
            if self.to_scheduler_param_list:
                for val_dict in self.to_scheduler_param_list:
                    self.set_scheduler(**val_dict)

    def _update_dict(self, source_dict: dict, target_dict: dict):
        for key, val in source_dict.items():
            target_dict[key] = val

    def save(self, save_path, global_step, name=None, default_save_param=None, *args, **kwargs):
        if default_save_param is None:
            default_save_param = dict(_parameters=self._parameters, _source_config=self._source_config.config_dict)
        if not name:
            name = self.name
        Logger().out_to_file(file_path=save_path,
                             file_name='{}-{}.json'.format(name, global_step),
                             content=default_save_param)

    def load(self, load_path, name, global_step, *args, **kwargs):
        res = files.load_json(file_path=os.path.join(load_path, "{}-{}.json".format(name, global_step)))
        # todo this mapping can be done via a dict structure
        if '_parameters' in res:
            setattr(self, '_parameters', res['_parameters'])
        if '_source_config' in res:
            setattr(self._source_config, 'config_dict', res['_source_config'])

    @typechecked
    def set_scheduler(self, param_key: str, scheduler: Scheduler, **kwargs):
        ori_value = self(param_key)
        scheduler.initial_p = ori_value
        self._scheduler_info_dict[param_key] = dict(param_key=param_key, scheduler=scheduler)

    def update(self, *args, **kwargs):
        for key, val in self._scheduler_info_dict.items():
            self.set(key=val['param_key'],
                     new_val=val['scheduler'].value())

    def set(self, key, new_val):
        if not isinstance(new_val, type(self(key))):
            raise TypeError('new value of parameters {} should be type {} instead of {}'.format(key, type(self(key)),
                                                                                                type(new_val)))
        elif key in self._parameters:
            self._parameters[key] = new_val
        else:
            self._source_config.set(key, new_val)



"""
This gives a simple example on how to use Gaussian Process (GP) to approximate the Gym environment Pendulum-v0
We use gpflow package to build the Gaussian Process.
"""
from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
import numpy as np
from baconian.common.sampler.sample_data import TransitionData
from baconian.algo.policy import UniformRandomPolicy
from baconian.algo.dynamics.gaussian_process_dynamiocs_model import GaussianProcessDyanmicsModel
from baconian.algo.dynamics.dynamics_model import DynamicsEnvWrapper
from baconian.algo.dynamics.terminal_func.terminal_func import RandomTerminalFunc
from baconian.algo.dynamics.reward_func.reward_func import RandomRewardFunc

env = make('Pendulum-v0')
name = 'demo_exp'
env_spec = EnvSpec(obs_space=env.observation_space,
                   action_space=env.action_space)
data = TransitionData(env_spec=env_spec)
policy = UniformRandomPolicy(env_spec=env_spec)
# Do some initial sampling here to train GP model
st = env.reset()
for i in range(100):
    ac = policy.forward(st)
    new_st, re, _, _ = env.step(ac)
    data.append(state=st, new_state=new_st, action=ac, reward=re, done=False)
    st = new_st

gp = GaussianProcessDyanmicsModel(env_spec=env_spec, batch_data=data)
gp.init()
gp.train()

dyna_env = DynamicsEnvWrapper(dynamics=gp)
# Since we only care about the prediction here, so we pass the terminal function and reward function setting with
# random one
dyna_env.set_terminal_reward_func(terminal_func=RandomTerminalFunc(),
                                  reward_func=RandomRewardFunc())

st = env.reset()
real_state_list = []
dynamics_state_list = []
test_sample_count = 100
for i in range(test_sample_count):
    ac = env_spec.action_space.sample()
    gp.reset_state(state=st)
    new_state_dynamics, _, _, _ = dyna_env.step(action=ac, allow_clip=True)
    new_state_real, _, done, _ = env.step(action=ac)
    real_state_list.append(new_state_real)
    dynamics_state_list.append(new_state_dynamics)
    st = new_state_real
    if done is True:
        env.reset()
l1_loss = np.linalg.norm(np.array(real_state_list) - np.array(dynamics_state_list), ord=1)
l2_loss = np.linalg.norm(np.array(real_state_list) - np.array(dynamics_state_list), ord=2)
print('l1 loss is {}, l2 loss is {}'.format(l1_loss, l2_loss))



from baconian.core.core import EnvSpec
from baconian.algo.dynamics.dynamics_model import DynamicsPriorModel
import numpy as np
from math import inf
from baconian.common.sampler.sample_data import TransitionData
from baconian.algo.dynamics.third_party.gmm import GMM
from baconian.core.parameters import Parameters


class GaussianMixtureDynamicsPrior(DynamicsPriorModel):
    """
    A dynamics prior encoded as a GMM over [x_t, u_t, x_t+1] points.
    See:
        S. Levine*, C. Finn*, T. Darrell, P. Abbeel, "End-to-end
        training of Deep Visuomotor Policies", arXiv:1504.00702,
        Appendix A.3.
    """

    def __init__(self, env_spec: EnvSpec, batch_data: TransitionData = None, epsilon=inf, init_sequential=False,
                 eigreg=False, warmstart=True, name_scope='gp_dynamics_model', min_samples_per_cluster=40,
                 max_clusters=20, strength=1,
                 name='gp_dynamics_model'):
        parameters = Parameters(
            dict(min_samp=min_samples_per_cluster, max_samples=inf, max_clusters=max_clusters, strength=strength,
                 init_sequential=init_sequential, eigreg=eigreg, warmstart=warmstart))
        super().__init__(env_spec=env_spec, parameters=parameters, name=name)
        self.name_scope = name_scope
        self.batch_data = batch_data
        self.gmm_model = GMM(epsilon=epsilon, init_sequential=init_sequential, eigreg=eigreg, warmstart=warmstart)
        self.X, self.U = None, None

    def init(self):
        pass

    def update(self, batch_data: TransitionData = None):
        """
        Update prior with additional data.
        :param batch_data: data used to update GMM prior
        :return: None
        """

        # Format Data
        xux, K = self._prepare_data(batch_data)

        # Update GMM.
        self.gmm_model.update(xux, K)

    def eval(self, batch_data: TransitionData = None):
        """
        Evaluate prior (prob of [x(t), u(t), x(t+1)] given gmm)

        :param batch_data: data used to evaluate the prior with.
        :return: parameters mu0, Phi, m, n0 as defined in the paper.
        """
        # Format Data
        xux, _ = self._prepare_data(batch_data)

        # Perform query and fix mean.
        mu0, Phi, m, n0 = self.gmm_model.inference(xux)

        # Factor in multiplier.
        n0 = n0 * self.parameters('strength')
        m = m * self.parameters('strength')

        # Multiply Phi by m (since it was normalized before).
        Phi *= m
        return mu0, Phi, m, n0

    def _prepare_data(self, batch_data: TransitionData = None, *kwargs):
        if self.batch_data is None:
            self.batch_data = batch_data

        X = batch_data.state_set
        if X.ndim == 2: X = np.expand_dims(X, axis=0)
        U = batch_data.action_set
        if U.ndim == 2: U = np.expand_dims(U, axis=0)

        # Constants.
        T = X.shape[1] - 1

        # Append data to dataset.
        if self.X is None:
            self.X = X
        else:
            self.X = np.concatenate([self.X, X], axis=0)

        if self.U is None:
            self.U = U
        else:
            self.U = np.concatenate([self.U, U], axis=0)

        # Remove excess samples from dataset.
        start = max(0, self.X.shape[0] - self.parameters('max_samples') + 1)
        self.X = self.X[start:, :]
        self.U = self.U[start:, :]

        # Compute cluster dimensionality.
        Do = X.shape[2] + U.shape[2] + X.shape[2]  # TODO: Use Xtgt.

        # Create dataset.
        N = self.X.shape[0]
        xux = np.reshape(
            np.c_[self.X[:, :T, :], self.U[:, :T, :], self.X[:, 1:(T + 1), :]],
            [T * N, Do]
        )

        # Choose number of clusters.
        K = int(max(2, min(self.parameters('max_clusters'), np.floor(float(N * T) / self.parameters('min_samp')))))

        return xux, K



MOUNTAIN_CAR_BENCHMARK_CONFIG_DICT = {
    'env_id': "MountainCarContinuous-v0",
    'MLPQValueFunction': {
        'mlp_config': [
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "1",
                "N_UNITS": 32,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "2",
                "N_UNITS": 64,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "3",
                "N_UNITS": 200,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "IDENTITY",
                "B_INIT_VALUE": 0.0,
                "NAME": "OUPTUT",
                "N_UNITS": 1,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            }
        ]
    },
    'DeterministicMLPPolicy': {
        'mlp_config': [
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "1",
                "N_UNITS": 8,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "2",
                "N_UNITS": 8,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "3",
                "N_UNITS": 8,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "IDENTITY",
                "B_INIT_VALUE": 0.0,
                "NAME": "OUPTUT",
                "N_UNITS": 1,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            }
        ]
    },
    'DDPG': {
        'config_or_config_dict': {
            "REPLAY_BUFFER_SIZE": 10000,
            "GAMMA": 0.99,
            "CRITIC_LEARNING_RATE": 0.001,
            "ACTOR_LEARNING_RATE": 0.0001,
            "DECAY": 0.5,
            "BATCH_SIZE": 128,
            "TRAIN_ITERATION": 120,
            "critic_clip_norm": None,
            "actor_clip_norm": None,
        },
        'replay_buffer': None
    },
    'EpsilonGreedy': {
        'initial_p': 1.0,
        'final_p': 0.0,
        'schedule_timesteps': 10000
    },
    'DEFAULT_EXPERIMENT_END_POINT': dict(TOTAL_ENV_STEP_TRAIN_SAMPLE_COUNT=10000),
    'DynaFlow': {
        "TEST_ALGO_EVERY_REAL_SAMPLE_COUNT": 100,
        "TEST_DYNAMICS_EVERY_REAL_SAMPLE_COUNT": 100,
        "TRAIN_DYNAMICS_EVERY_REAL_SAMPLE_COUNT": 20,
        "START_TRAIN_ALGO_AFTER_SAMPLE_COUNT": 1,
        "START_TRAIN_DYNAMICS_AFTER_SAMPLE_COUNT": 1,
        "START_TEST_ALGO_AFTER_SAMPLE_COUNT": 1,
        "START_TEST_DYNAMICS_AFTER_SAMPLE_COUNT": 1,
        "WARM_UP_DYNAMICS_SAMPLES": 10000,
        "TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_REAL_ENV": 20,
        "TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_DYNAMICS_ENV": 20,
    },
    'DynamicsModel': dict(learning_rate=0.01,
                          mlp_config=[
                              {
                                  "ACT": "RELU",
                                  "B_INIT_VALUE": 0.0,
                                  "NAME": "1",
                                  "N_UNITS": 32,
                                  "TYPE": "DENSE",
                                  "W_NORMAL_STDDEV": 0.03
                              },
                              {
                                  "ACT": "RELU",
                                  "B_INIT_VALUE": 0.0,
                                  "NAME": "2",
                                  "N_UNITS": 64,
                                  "TYPE": "DENSE",
                                  "W_NORMAL_STDDEV": 0.03
                              },
                              {
                                  "ACT": "IDENTITY",
                                  "B_INIT_VALUE": 0.0,
                                  "NAME": "OUPTUT",
                                  "N_UNITS": 3,
                                  "TYPE": "DENSE",
                                  "W_NORMAL_STDDEV": 0.03
                              }
                          ])

}



"""
This script show the example for adding an early stopping feature so when the agent can't increase its received average
reward for evaluation, the experiment will end early.

To do so in a extensible and modular way. We can implement a new flow called EarlyStoppingFlow that implement a special
ending condition detections by accessing the agent's evaluation reward (with built-in modular to access). Such mechanism
can be re-used by all algorithms, which avoid the redundant coding for users.
"""
from baconian.config.dict_config import DictConfig
from baconian.core.flow.train_test_flow import TrainTestFlow

from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.algo.value_func.mlp_q_value import MLPQValueFunction
from baconian.algo.ddpg import DDPG
from baconian.algo.policy import DeterministicMLPPolicy
from baconian.core.agent import Agent
from baconian.algo.misc import EpsilonGreedy
from baconian.core.experiment import Experiment
from baconian.core.status import get_global_status_collect
from baconian.common.schedules import PeriodicalEventSchedule


class EarlyStoppingFlow(TrainTestFlow):
    required_key_dict = {
        **TrainTestFlow.required_key_dict,
        'USE_LAST_K_EVALUATION_REWARD': 10
    }

    def __init__(self, train_sample_count_func, config_or_config_dict: (DictConfig, dict), func_dict: dict, agent):
        super().__init__(train_sample_count_func, config_or_config_dict, func_dict)
        self.agent = agent

    def _is_ended(self):
        test_reward = sorted(self.agent.recorder.get_log(attr_name='sum_reward', filter_by_status=dict(status='TEST')),
                             key=lambda x: x['sample_counter'])
        if len(test_reward) >= self.parameters('USE_LAST_K_EVALUATION_REWARD') * 2:
            last_reward = test_reward[-self.parameters('USE_LAST_K_EVALUATION_REWARD'):]
            pre_reward = test_reward[-self.parameters('USE_LAST_K_EVALUATION_REWARD') * 2: -self.parameters(
                'USE_LAST_K_EVALUATION_REWARD')]
            last_reward = np.mean([r['log_val'] for r in last_reward])
            pre_reward = np.mean([r['log_val'] for r in pre_reward])
            if last_reward < pre_reward:
                ConsoleLogger().print('info', 'training ended because last {} step reward: {} < previous {} step reward {}'.format(self.parameters('USE_LAST_K_EVALUATION_REWARD'), last_reward, self.parameters('USE_LAST_K_EVALUATION_REWARD'), pre_reward))
                return True
        return super()._is_ended()


def create_early_stopping_flow(test_every_sample_count, train_every_sample_count, start_train_after_sample_count,
                               start_test_after_sample_count, train_func_and_args, test_func_and_args,
                               sample_func_and_args,
                               agent,
                               use_last_k_evaluation_reward,
                               train_samples_counter_func=None):
    config_dict = dict(
        TEST_EVERY_SAMPLE_COUNT=test_every_sample_count,
        TRAIN_EVERY_SAMPLE_COUNT=train_every_sample_count,
        START_TRAIN_AFTER_SAMPLE_COUNT=start_train_after_sample_count,
        START_TEST_AFTER_SAMPLE_COUNT=start_test_after_sample_count,
        USE_LAST_K_EVALUATION_REWARD=use_last_k_evaluation_reward
    )

    def return_func_dict(s_dict):
        return dict(func=s_dict[0],
                    args=s_dict[1],
                    kwargs=s_dict[2])

    func_dict = dict(
        train=return_func_dict(train_func_and_args),
        test=return_func_dict(test_func_and_args),
        sample=return_func_dict(sample_func_and_args),
    )
    if train_samples_counter_func is None:
        def default_train_samples_counter_func():
            return get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT')

        train_samples_counter_func = default_train_samples_counter_func

    return EarlyStoppingFlow(config_or_config_dict=config_dict,
                             train_sample_count_func=train_samples_counter_func,
                             agent=agent,
                             func_dict=func_dict)


def task_fn():
    env = make('Pendulum-v0')
    name = 'demo_exp'
    env_spec = EnvSpec(obs_space=env.observation_space,
                       action_space=env.action_space)

    mlp_q = MLPQValueFunction(env_spec=env_spec,
                              name_scope=name + '_mlp_q',
                              name=name + '_mlp_q',
                              mlp_config=[
                                  {
                                      "ACT": "RELU",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "1",
                                      "N_UNITS": 16,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  },
                                  {
                                      "ACT": "LINEAR",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "OUPTUT",
                                      "N_UNITS": 1,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  }
                              ])
    policy = DeterministicMLPPolicy(env_spec=env_spec,
                                    name_scope=name + '_mlp_policy',
                                    name=name + '_mlp_policy',
                                    mlp_config=[
                                        {
                                            "ACT": "RELU",
                                            "B_INIT_VALUE": 0.0,
                                            "NAME": "1",
                                            "N_UNITS": 16,
                                            "TYPE": "DENSE",
                                            "W_NORMAL_STDDEV": 0.03
                                        },
                                        {
                                            "ACT": "LINEAR",
                                            "B_INIT_VALUE": 0.0,
                                            "NAME": "OUPTUT",
                                            "N_UNITS": env_spec.flat_action_dim,
                                            "TYPE": "DENSE",
                                            "W_NORMAL_STDDEV": 0.03
                                        }
                                    ],
                                    reuse=False)

    ddpg = DDPG(
        env_spec=env_spec,
        config_or_config_dict={
            "REPLAY_BUFFER_SIZE": 10000,
            "GAMMA": 0.999,
            "CRITIC_LEARNING_RATE": 0.001,
            "ACTOR_LEARNING_RATE": 0.001,
            "DECAY": 0.5,
            "BATCH_SIZE": 50,
            "TRAIN_ITERATION": 1,
            "critic_clip_norm": 0.1,
            "actor_clip_norm": 0.1,
        },
        value_func=mlp_q,
        policy=policy,
        name=name + '_ddpg',
        replay_buffer=None
    )
    agent = Agent(env=env, env_spec=env_spec,
                  algo=ddpg,
                  algo_saving_scheduler=PeriodicalEventSchedule(
                      t_fn=lambda: get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
                      trigger_every_step=20,
                      after_t=10),
                  name=name + '_agent',
                  exploration_strategy=EpsilonGreedy(action_space=env_spec.action_space,
                                                     init_random_prob=0.5))

    flow = create_early_stopping_flow(
        agent=agent,
        use_last_k_evaluation_reward=5,
        test_every_sample_count=10,
        train_every_sample_count=10,
        start_test_after_sample_count=5,
        start_train_after_sample_count=5,
        train_func_and_args=(agent.train, (), dict()),
        test_func_and_args=(agent.test, (), dict(sample_count=1)),
        sample_func_and_args=(agent.sample, (), dict(sample_count=100,
                                                     env=agent.env))
    )

    experiment = Experiment(
        tuner=None,
        env=env,
        agent=agent,
        flow=flow,
        name=name
    )
    experiment.run()


from baconian.core.experiment_runner import *

GlobalConfig().set('DEFAULT_LOG_PATH', './log_path')
GlobalConfig().set('DEFAULT_EXPERIMENT_END_POINT',
                   dict(TOTAL_AGENT_TRAIN_SAMPLE_COUNT=2000,
                        TOTAL_AGENT_TEST_SAMPLE_COUNT=None,
                        TOTAL_AGENT_UPDATE_COUNT=None))
single_exp_runner(task_fn, del_if_log_path_existed=True)



class BaconianError(Exception):
    pass


class GlobalNameExistedError(BaconianError):
    pass


class StatusInfoNotRegisteredError(BaconianError):
    pass


class StateOrActionOutOfBoundError(BaconianError):
    pass


class MemoryBufferLessThanBatchSizeError(BaconianError):
    pass


class InappropriateParameterSetting(BaconianError):
    pass


class DuplicatedRegisteredError(BaconianError):
    pass


class LogPathOrFileNotExistedError(BaconianError):
    pass


class NotCatchCorrectExceptionError(BaconianError):
    pass


class AttemptToChangeFreezeGlobalConfigError(BaconianError):
    pass


class MissedConfigError(BaconianError):
    pass


class TransformationResultedToDifferentShapeError(BaconianError):
    pass


class WrongValueRangeError(BaconianError):
    pass


class ShapeNotCompatibleError(BaconianError):
    pass


class EnvNotExistedError(BaconianError):
    pass


class LogItemNotExisted(BaconianError):
    pass



import unittest
import numpy as np
from baconian.tf.util import MLPCreator
from baconian.test.tests.set_up.setup import TestTensorflowSetup
import tensorflow as tf


class TestTensorflowUtil(TestTensorflowSetup):

    def test_init_with_tf_layers(self):
        input_ph = tf.placeholder(dtype=tf.float32, shape=[None, 5], name='ph1')
        input_ph2 = tf.placeholder(dtype=tf.float32, shape=[None, 5], name='ph1')
        net1_name = 'net'
        net1_scope = 'scope'

        net2_name = 'net'
        net2_scope = 'scope'

        net1 = MLPCreator.create_network_with_tf_layers(input=input_ph,
                                                        network_config=[
                                                            {
                                                                "ACT": "RELU",
                                                                "B_INIT_VALUE": 0.0,
                                                                "NAME": "1",
                                                                "N_UNITS": 16,
                                                                "TYPE": "DENSE",
                                                                "W_NORMAL_STDDEV": 0.03
                                                            },
                                                            {
                                                                "ACT": "LINEAR",
                                                                "B_INIT_VALUE": 0.0,
                                                                "NAME": "OUPTUT",
                                                                "N_UNITS": 1,
                                                                "TYPE": "DENSE",
                                                                "W_NORMAL_STDDEV": 0.03
                                                            }
                                                        ],
                                                        net_name=net1_name,
                                                        reuse=False,
                                                        tf_var_scope=net1_scope)

        net2 = MLPCreator.create_network_with_tf_layers(input=input_ph,
                                                        network_config=[
                                                            {
                                                                "ACT": "RELU",
                                                                "B_INIT_VALUE": 0.0,
                                                                "NAME": "1",
                                                                "N_UNITS": 16,
                                                                "TYPE": "DENSE",
                                                                "W_NORMAL_STDDEV": 0.03
                                                            },
                                                            {
                                                                "ACT": "LINEAR",
                                                                "B_INIT_VALUE": 0.0,
                                                                "NAME": "OUPTUT",
                                                                "N_UNITS": 1,
                                                                "TYPE": "DENSE",
                                                                "W_NORMAL_STDDEV": 0.03
                                                            }
                                                        ],
                                                        net_name=net2_name,
                                                        reuse=True,
                                                        tf_var_scope=net2_scope)
        self.assertGreater(len(net1[2]), 0)
        self.assertGreater(len(net2[2]), 0)
        self.sess.run(tf.global_variables_initializer())
        for var1, var2 in zip(net1[2], net2[2]):
            print("net1: {} {}  | true name: {}".format(net1_scope, net1_name, var1.name))
            print("net2: {} {} | true name: {}".format(net2_scope, net2_name, var2.name))
            self.assertEqual(id(var1), id(var2))
            self.assertTrue(net1_name in var1.name)
            self.assertTrue(net1_scope in var1.name)
            self.assertTrue(net2_name in var2.name)
            self.assertTrue(net2_scope in var2.name)

        var = net1[2][0]
        op = tf.assign(var,
                       tf.constant(value=np.random.random(list(self.sess.run(tf.shape(var)))),
                                   dtype=tf.float32))
        self.sess.run(op)
        var1 = self.sess.run(var)
        var2 = self.sess.run(net2[2][0])
        self.assertTrue(np.equal(var1, var2).all())

    def test_init_with_tf_layers_2(self):
        input_ph = tf.placeholder(dtype=tf.float32, shape=[None, 5], name='ph1')
        net1_name = 'net'
        net1_scope = 'scope'

        net2_name = 'net'
        net2_scope = 'scope1'

        net1 = MLPCreator.create_network_with_tf_layers(input=input_ph,
                                                        network_config=[
                                                            {
                                                                "ACT": "RELU",
                                                                "B_INIT_VALUE": 0.0,
                                                                "NAME": "1",
                                                                "N_UNITS": 16,
                                                                "TYPE": "DENSE",
                                                                "W_NORMAL_STDDEV": 0.03
                                                            },
                                                            {
                                                                "ACT": "LINEAR",
                                                                "B_INIT_VALUE": 0.0,
                                                                "NAME": "OUPTUT",
                                                                "N_UNITS": 1,
                                                                "TYPE": "DENSE",
                                                                "W_NORMAL_STDDEV": 0.03
                                                            }
                                                        ],
                                                        net_name=net1_name,
                                                        reuse=False,
                                                        tf_var_scope=net1_scope)

        net2 = MLPCreator.create_network_with_tf_layers(input=input_ph,
                                                        network_config=[
                                                            {
                                                                "ACT": "RELU",
                                                                "B_INIT_VALUE": 0.0,
                                                                "NAME": "1",
                                                                "N_UNITS": 16,
                                                                "TYPE": "DENSE",
                                                                "W_NORMAL_STDDEV": 0.03
                                                            },
                                                            {
                                                                "ACT": "LINEAR",
                                                                "B_INIT_VALUE": 0.0,
                                                                "NAME": "OUPTUT",
                                                                "N_UNITS": 1,
                                                                "TYPE": "DENSE",
                                                                "W_NORMAL_STDDEV": 0.03
                                                            }
                                                        ],
                                                        net_name=net2_name,
                                                        reuse=False,
                                                        tf_var_scope=net2_scope)
        self.assertGreater(len(net1[2]), 0)
        self.assertGreater(len(net2[2]), 0)
        self.sess.run(tf.global_variables_initializer())
        for var1, var2 in zip(net1[2], net2[2]):
            print("net1: {} {}  | true name: {}".format(net1_scope, net1_name, var1.name))
            print("net2: {} {} | true name: {}".format(net2_scope, net2_name, var2.name))
            self.assertFalse(id(var1) == id(var2))
            self.assertTrue(net1_name in var1.name)
            self.assertTrue(net1_scope in var1.name)
            self.assertTrue(net2_name in var2.name)
            self.assertTrue(net2_scope in var2.name)
        var = net1[2][0]
        op = tf.assign(var,
                       tf.constant(value=np.random.random(list(self.sess.run(tf.shape(var)))),
                                   dtype=tf.float32))
        self.sess.run(op)
        var1 = self.sess.run(var)
        var2 = self.sess.run(net2[2][0])
        self.assertFalse(np.equal(var1, var2).all())

    def test_init_with_tf_layers_3(self):
        input_ph = tf.placeholder(dtype=tf.float32, shape=[None, 5], name='ph1')
        net1_name = 'net'
        net1_scope = 'scope'

        net2_name = 'net'
        net2_scope = 'scope'

        net1 = MLPCreator.create_network_with_tf_layers(input=input_ph,
                                                        network_config=[
                                                            {
                                                                "ACT": "RELU",
                                                                "B_INIT_VALUE": 0.0,
                                                                "NAME": "1",
                                                                "N_UNITS": 16,
                                                                "TYPE": "DENSE",
                                                                "W_NORMAL_STDDEV": 0.03
                                                            },
                                                            {
                                                                "ACT": "LINEAR",
                                                                "B_INIT_VALUE": 0.0,
                                                                "NAME": "OUPTUT",
                                                                "N_UNITS": 1,
                                                                "TYPE": "DENSE",
                                                                "W_NORMAL_STDDEV": 0.03
                                                            }
                                                        ],
                                                        net_name=net1_name,
                                                        reuse=False,
                                                        tf_var_scope=net1_scope)
        input_ph2 = tf.placeholder(dtype=tf.float32, shape=[None, 5], name='ph1')
        input_ph2 = tf.tanh(input_ph2)
        net2 = MLPCreator.create_network_with_tf_layers(input=input_ph2,
                                                        network_config=[
                                                            {
                                                                "ACT": "RELU",
                                                                "B_INIT_VALUE": 0.0,
                                                                "NAME": "1",
                                                                "N_UNITS": 16,
                                                                "TYPE": "DENSE",
                                                                "W_NORMAL_STDDEV": 0.03
                                                            },
                                                            {
                                                                "ACT": "LINEAR",
                                                                "B_INIT_VALUE": 0.0,
                                                                "NAME": "OUPTUT",
                                                                "N_UNITS": 1,
                                                                "TYPE": "DENSE",
                                                                "W_NORMAL_STDDEV": 0.03
                                                            }
                                                        ],
                                                        net_name=net2_name,
                                                        reuse=True,
                                                        tf_var_scope=net2_scope)
        self.assertGreater(len(net1[2]), 0)
        self.assertGreater(len(net2[2]), 0)
        self.assertEqual(len(net1), len(net2))
        self.sess.run(tf.global_variables_initializer())
        for var1, var2 in zip(net1[2], net2[2]):
            print("net1: {} {}  | true name: {}".format(net1_scope, net1_name, var1.name))
            print("net2: {} {} | true name: {}".format(net2_scope, net2_name, var2.name))
            self.assertTrue(id(var1) == id(var2))
            self.assertTrue(net1_name in var1.name)
            self.assertTrue(net1_scope in var1.name)
            self.assertTrue(net2_name in var2.name)
            self.assertTrue(net2_scope in var2.name)
        var = net1[2][0]
        op = tf.assign(var,
                       tf.constant(value=np.random.random(list(self.sess.run(tf.shape(var)))),
                                   dtype=tf.float32))
        self.sess.run(op)
        var1 = self.sess.run(var)
        var2 = self.sess.run(net2[2][0])
        self.assertTrue(np.equal(var1, var2).all())


if __name__ == '__main__':
    unittest.main()



"""This is a garage-compatible wrapper for Dict spaces."""
from collections import OrderedDict

from baconian.common.spaces.base import Space
from baconian.common.spaces import Box

import numpy as np
import types


class Dict(Space):
    """
    A dictionary of simpler spaces, e.g. Discrete, Box.

    Example usage:
        self.observation_space = spaces.Dict({"position": spaces.Discrete(2),
                                              "velocity": spaces.Discrete(3)})
    """

    def __init__(self, spaces):
        """
        Convert and store the incoming spaces into an OrderedDict.

        Note: classes inheriting from garage.Dict need to convert each
        space in spaces to a garage.<class>.space.
        """
        if isinstance(spaces, dict):
            spaces = OrderedDict(sorted(list(spaces.items())))
        if isinstance(spaces, list):
            spaces = OrderedDict(spaces)
        self.spaces = spaces

    def contains(self, x):
        """
        Check if x is contained within self.spaces.

        Returns:
            Boolean

        """
        if isinstance(x, dict):
            return bool(all(item in self.spaces.items() for item in x.items()))
        else:
            return False

    def to_jsonable(self, sample_n):
        """
        Serialize as a dict-representation of vectors.

        Returns:
            JSON (dict)

        """
        return {key: space.to_jsonable([sample[key] for sample in sample_n]) \
                for key, space in self.spaces.items()}

    def from_jsonable(self, sample_n):
        """
        Convert information from a JSON format into a list.

        Returns:
            ret (list)

        """
        dict_of_list = {}
        for key, space in self.spaces.items():
            dict_of_list[key] = space.from_jsonable(sample_n[key])
        ret = []
        for i, _ in enumerate(dict_of_list[key]):
            entry = {}
            for key, value in dict_of_list.items():
                entry[key] = value[i]
            ret.append(entry)
        return ret

    @property
    def flat_dim(self):
        """
        Return a flat dimension of the dict space.

        Returns:
            flat_dim (int)

        """
        raise NotImplementedError

    def flatten(self, x):
        """
        Return a flattened observation x.

        Returns:
            x (flattened)

        """
        raise NotImplementedError

    def unflatten(self, x):
        """
        Return an unflattened observation x.

        Returns:
            x (unflattened)

        """
        raise NotImplementedError

    def flatten_n(self, xs):
        """
        Return flattened observations xs.

        Returns:
            xs (flattened)

        """
        raise NotImplementedError

    def unflatten_n(self, xs):
        """
        Return unflattened observations xs.

        Returns:
            xs (unflattened)

        """
        raise NotImplementedError

    def sample(self):
        """
        Return a sample from each space in spaces.

        Returns:
            OrderedDict

        """
        # raise NotImplementedError
        # return OrderedDict([(k, space.sample()) for k, space in self.spaces.items()])

        ordered = OrderedDict()
        for k, space in self.spaces.items():
            for a in space.low:
                if np.isinf(a):
                    a = np.nan_to_num(a)
                    space.sample = types.MethodType(self._sample_with_nan, space)
            for b in space.high:
                if np.isinf(b):
                    b = np.nan_to_num(b)
                    space.sample = types.MethodType(self._sample_with_nan, space)

            ordered.update([(k, space.sample())])
        return ordered

    def new_tensor_variable(self, name, extra_dims):
        """
        Return a new tensor variable in the TF graph.

        Returns:
            Tensor

        """
        raise NotImplementedError

    @staticmethod
    def _sample_with_nan(space: Space):
        """

        :param space:
        :return:
        """
        from gym.spaces.box import Box as GymBox

        if not isinstance(space, GymBox):
            raise TypeError('space is not of type Box')
        high = np.ones_like(space.low)
        low = -1 * np.ones_like(space.high)
        return np.clip(np.random.uniform(low=low, high=high, size=space.low.shape),
                       a_min=space.low,
                       a_max=space.high)



from baconian.envs.gym_env import make
from baconian.core.core import EnvSpec
from baconian.test.tests.set_up.setup import BaseTestCase
from baconian.algo.misc.replay_buffer import UniformRandomReplayBuffer


class TestReplaybuffer(BaseTestCase):
    def test_transition_data(self):
        env = make('Acrobot-v1')
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)
        a = UniformRandomReplayBuffer(limit=10000, action_shape=env_spec.action_shape,
                                      observation_shape=env_spec.obs_shape)
        st = env.reset()
        for i in range(100):
            ac = env_spec.action_space.sample()
            st_new, re, done, _ = env.step(action=ac)
            a.append(obs0=st, obs1=st_new, action=ac, reward=re, terminal1=done)
            st = st_new
        batch = a.sample(batch_size=10)
        self.assertTrue(batch.state_set.shape[0] == 10)
        self.assertTrue(batch.action_set.shape[0] == 10)
        self.assertTrue(batch.reward_set.shape[0] == 10)
        self.assertTrue(batch.done_set.shape[0] == 10)
        self.assertTrue(batch.new_state_set.shape[0] == 10)



import abc
from baconian.config.global_config import GlobalConfig
from baconian.common.logging import ConsoleLogger
from baconian.config.dict_config import DictConfig
from baconian.common.misc import *
from baconian.core.parameters import Parameters
from baconian.core.status import *
from baconian.common.error import *


class Flow(object):
    """
    Interface of experiment flow module, it defines the workflow of the reinforcement learning experiments.
    """
    required_func = ()
    required_key_dict = dict()

    def __init__(self, func_dict):
        """
        Constructor for Flow.

        :param func_dict: the function and its arguments that will be called in the Flow
        :type func_dict: dict
        """
        self.func_dict = func_dict
        for key in self.required_func:
            if key not in func_dict:
                raise MissedConfigError('miss key {}'.format(key))

    def launch(self) -> bool:
        """
        Launch the flow until it finished or catch a system-allowed errors (e.g., out of GPU memory, to ensure the log will be saved safely).

        :return: True if the flow correctly executed and finished
        :rtype: bool
        """
        try:
            return self._launch()
        except GlobalConfig().DEFAULT_ALLOWED_EXCEPTION_OR_ERROR_LIST as e:
            ConsoleLogger().print('error', 'error {} occurred'.format(e))
            return False

    def _launch(self) -> bool:
        """
        Abstract method to be implemented by subclass for a certain workflow.

        :return: True if the flow correctly executed and finished
        :rtype: bool
        """
        raise NotImplementedError

    def _call_func(self, key, **extra_kwargs):
        """
        Call a function that is pre-defined in self.func_dict

        :param key: name of the function, e.g., train, test, sample.
        :type key: str
        :param extra_kwargs: some extra kwargs you may want to be passed in the function calling
        :return: actual return value of the called function if self.func_dict has such function otherwise None.
        :rtype:
        """

        if self.func_dict[key]:
            return self.func_dict[key]['func'](*self.func_dict[key]['args'],
                                               **extra_kwargs,
                                               **self.func_dict[key]['kwargs'])
        else:
            return None


class TrainTestFlow(Flow):
    """
    A typical sampling-trainning and testing workflow, that used by most of model-free/model-based reinforcement
    learning method. Typically, it repeat the sampling(saving to memory if off policy)->training(from memory if
    off-policy, from samples if on-policy)->test
    """
    required_func = ('train', 'test', 'sample')
    required_key_dict = {
        "TEST_EVERY_SAMPLE_COUNT": 1000,
        "TRAIN_EVERY_SAMPLE_COUNT": 1000,
        "START_TRAIN_AFTER_SAMPLE_COUNT": 1,
        "START_TEST_AFTER_SAMPLE_COUNT": 1,
    }

    def __init__(self,
                 train_sample_count_func,
                 config_or_config_dict: (DictConfig, dict),
                 func_dict: dict,
                 ):
        """
        Constructor of TrainTestFlow

        :param train_sample_count_func: a function indicates how much training samples the agent has collected currently.
        :type train_sample_count_func: method
        :param config_or_config_dict: a Config or a dict should have the keys: (TEST_EVERY_SAMPLE_COUNT, TRAIN_EVERY_SAMPLE_COUNT, START_TRAIN_AFTER_SAMPLE_COUNT, START_TEST_AFTER_SAMPLE_COUNT)
        :type config_or_config_dict: Config or dict
        :param func_dict: function dict, holds the keys: 'sample', 'train', 'test'. each item in the dict as also should be a dict, holds the keys 'func', 'args', 'kwargs'
        :type func_dict: dict
        """
        super(TrainTestFlow, self).__init__(func_dict=func_dict)
        config = construct_dict_config(config_or_config_dict, obj=self)
        self.parameters = Parameters(source_config=config, parameters=dict())
        self.time_step_func = train_sample_count_func
        self.last_train_point = -1
        self.last_test_point = -1
        assert callable(train_sample_count_func)

    def _launch(self) -> bool:
        """
        Launch the flow until it finished or catch a system-allowed errors
        (e.g., out of GPU memory, to ensure the log will be saved safely).

        :return: True if the flow correctly executed and finished
        :rtype: bool
        """
        while True:
            self._call_func('sample')
            if self.time_step_func() - self.parameters('TRAIN_EVERY_SAMPLE_COUNT') >= self.last_train_point and \
                    self.time_step_func() > self.parameters('START_TRAIN_AFTER_SAMPLE_COUNT'):
                self.last_train_point = self.time_step_func()
                self._call_func('train')
            if self.time_step_func() - self.parameters('TEST_EVERY_SAMPLE_COUNT') >= self.last_test_point and \
                    self.time_step_func() > self.parameters('START_TEST_AFTER_SAMPLE_COUNT'):
                self.last_test_point = self.time_step_func()
                self._call_func('test')

            if self._is_ended() is True:
                break
        return True

    def _is_ended(self):
        """

        :return: True if an experiment is ended
        :rtype: bool
        """
        key_founded_flag = False
        finished_flag = False
        for key in GlobalConfig().DEFAULT_EXPERIMENT_END_POINT:
            if GlobalConfig().DEFAULT_EXPERIMENT_END_POINT[key] is not None:
                key_founded_flag = True
                if get_global_status_collect()(key) >= GlobalConfig().DEFAULT_EXPERIMENT_END_POINT[key]:
                    ConsoleLogger().print('info',
                                          'pipeline ended because {}: {} >= end point value {}'.
                                          format(key, get_global_status_collect()(key),
                                                 GlobalConfig().DEFAULT_EXPERIMENT_END_POINT[key]))
                    finished_flag = True
        if key_founded_flag is False:
            ConsoleLogger().print(
                'warning',
                '{} in experiment_end_point is not registered with global status collector: {}, experiment may not end'.
                    format(GlobalConfig().DEFAULT_EXPERIMENT_END_POINT, list(get_global_status_collect()().keys())))
        return finished_flag


def create_train_test_flow(test_every_sample_count, train_every_sample_count, start_train_after_sample_count,
                           start_test_after_sample_count, train_func_and_args, test_func_and_args, sample_func_and_args,
                           train_samples_counter_func=None):
    config_dict = dict(
        TEST_EVERY_SAMPLE_COUNT=test_every_sample_count,
        TRAIN_EVERY_SAMPLE_COUNT=train_every_sample_count,
        START_TRAIN_AFTER_SAMPLE_COUNT=start_train_after_sample_count,
        START_TEST_AFTER_SAMPLE_COUNT=start_test_after_sample_count,
    )

    def return_func_dict(s_dict):
        return dict(func=s_dict[0],
                    args=s_dict[1],
                    kwargs=s_dict[2])

    func_dict = dict(
        train=return_func_dict(train_func_and_args),
        test=return_func_dict(test_func_and_args),
        sample=return_func_dict(sample_func_and_args),
    )
    if train_samples_counter_func is None:
        def default_train_samples_counter_func():
            return get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT')

        train_samples_counter_func = default_train_samples_counter_func

    return TrainTestFlow(config_or_config_dict=config_dict,
                         train_sample_count_func=train_samples_counter_func,
                         func_dict=func_dict)



from baconian.common.sampler.sample_data import TransitionData
from baconian.test.tests.set_up.setup import TestWithAll
from baconian.config.global_config import GlobalConfig


class TestDDPG(TestWithAll):
    def test_init(self):
        ddpg, locals = self.create_ddpg()
        self.assert_var_list_id_no_equal(var_list1=ddpg.actor.parameters('tf_var_list'),
                                         var_list2=ddpg.target_actor.parameters('tf_var_list'))

        self.assert_var_list_id_no_equal(var_list1=ddpg.critic.parameters('tf_var_list'),
                                         var_list2=ddpg.target_critic.parameters('tf_var_list'))

        self.assert_var_list_id_equal(var_list1=ddpg.critic.parameters('tf_var_list'),
                                      var_list2=ddpg._critic_with_actor_output.parameters('tf_var_list'))
        self.assert_var_list_id_equal(var_list1=ddpg.target_critic.parameters('tf_var_list'),
                                      var_list2=ddpg._target_critic_with_target_actor_output.parameters('tf_var_list'))

        env = locals['env']
        env_spec = locals['env_spec']
        ddpg.init()
        data = TransitionData(env_spec)
        st = env.reset()
        for i in range(100):
            ac = ddpg.predict(st)
            new_st, re, done, _ = env.step(ac)
            data.append(state=st, new_state=new_st, action=ac, reward=re, done=done)
            st = new_st
        ddpg.append_to_memory(data)
        new_ddpg, _ = self.create_ddpg(name='new_ddpg')
        new_ddpg.copy_from(ddpg)
        self.assert_var_list_equal(ddpg.actor.parameters('tf_var_list'),
                                   new_ddpg.actor.parameters('tf_var_list'))
        self.assert_var_list_equal(ddpg.critic.parameters('tf_var_list'),
                                   new_ddpg.critic.parameters('tf_var_list'))
        self.assert_var_list_equal(ddpg.target_actor.parameters('tf_var_list'),
                                   new_ddpg.target_actor.parameters('tf_var_list'))
        self.assert_var_list_equal(ddpg.target_critic.parameters('tf_var_list'),
                                   new_ddpg.target_critic.parameters('tf_var_list'))
        ddpg.save(save_path=GlobalConfig().DEFAULT_LOG_PATH + '/ddpg_test',
                  global_step=0,
                  name=ddpg.name)
        for i in range(100):
            print(ddpg.train(train_iter=10))
        self.assert_var_list_at_least_not_equal(ddpg.critic.parameters('tf_var_list'),
                                                new_ddpg.critic.parameters('tf_var_list'))
        self.assert_var_list_at_least_not_equal(ddpg.target_critic.parameters('tf_var_list'),
                                                new_ddpg.target_critic.parameters('tf_var_list'))
        self.assert_var_list_at_least_not_equal(ddpg.actor.parameters('tf_var_list'),
                                                new_ddpg.actor.parameters('tf_var_list'))

        self.assert_var_list_at_least_not_equal(ddpg.target_actor.parameters('tf_var_list'),
                                                new_ddpg.target_actor.parameters('tf_var_list'))

        ddpg.load(path_to_model=GlobalConfig().DEFAULT_LOG_PATH + '/ddpg_test',
                  model_name=ddpg.name,
                  global_step=0)

        self.assert_var_list_equal(ddpg.actor.parameters('tf_var_list'),
                                   new_ddpg.actor.parameters('tf_var_list'))



from baconian.core.core import EnvSpec
from baconian.algo.policy.policy import Policy
from typeguard import typechecked
from overrides.overrides import overrides
import numpy as np


class UniformRandomPolicy(Policy):

    @typechecked
    def __init__(self, env_spec: EnvSpec, name: str = 'random_policy'):
        super().__init__(env_spec=env_spec, name=name)

    @overrides
    def forward(self, obs, **kwargs):
        return np.array(self.action_space.sample())

    def save(self, global_step, save_path=None, name=None, **kwargs):
        pass

    def load(self, path_to_model, model_name, global_step=None, **kwargs):
        pass



from baconian.algo.algo import Algo
from baconian.algo.dynamics.dynamics_model import DynamicsModel
from baconian.core.core import EnvSpec
from baconian.common.logging import record_return_decorator
import numpy as np


class ModelFreeAlgo(Algo):
    def __init__(self, env_spec: EnvSpec, name: str = 'model_free_algo', warm_up_trajectories_number=0):
        super(ModelFreeAlgo, self).__init__(env_spec, name, warm_up_trajectories_number)


class OnPolicyAlgo(Algo):
    pass


class OffPolicyAlgo(Algo):
    pass


class ValueBasedAlgo(Algo):
    pass


class PolicyBasedAlgo(Algo):
    pass


class ModelBasedAlgo(Algo):
    def __init__(self, env_spec, dynamics_model: DynamicsModel, name: str = 'model_based_algo'):
        super(ModelBasedAlgo, self).__init__(env_spec, name)
        self._dynamics_model = dynamics_model
        self.dynamics_env = self._dynamics_model.return_as_env()

    def train_dynamics(self, *args, **kwargs):
        pass

    @record_return_decorator(which_recorder='self')
    def test_dynamics(self, env, sample_count, *args, **kwargs):
        self.set_status('TEST')
        env.set_status('TEST')
        st = env.reset()
        real_state_list = []
        dyanmics_state_list = []
        for i in range(sample_count):
            ac = self.env_spec.action_space.sample()
            self._dynamics_model.reset_state(state=st)
            new_state_dynamics, _, _, _ = self.dynamics_env.step(action=ac, )
            new_state_real, _, done, _ = env.step(action=ac)
            real_state_list.append(new_state_real)
            dyanmics_state_list.append(new_state_dynamics)
            st = new_state_real
            if done is True:
                env.reset()
        l1_loss = np.linalg.norm(np.array(real_state_list) - np.array(dyanmics_state_list), ord=1)
        l2_loss = np.linalg.norm(np.array(real_state_list) - np.array(dyanmics_state_list), ord=2)
        return dict(dynamics_test_l1_error=l1_loss, dynamics_test_l2_error=l2_loss)

    def set_terminal_reward_function_for_dynamics_env(self, terminal_func, reward_func):
        self.dynamics_env.set_terminal_reward_func(terminal_func, reward_func)



from baconian.common.sampler.sample_data import TransitionData
import unittest
from baconian.test.tests.set_up.setup import TestTensorflowSetup


class TestMPC(TestTensorflowSetup):

    def test_init_discrete(self):
        algo, locals = self.create_mpc()
        env_spec = locals['env_spec']
        env = locals['env']
        algo.init()
        for _ in range(100):
            assert env_spec.action_space.contains(algo.predict(env_spec.obs_space.sample()))

        st = env.reset()
        data = TransitionData(env_spec)

        for _ in range(10):
            ac = algo.predict(st)
            new_st, re, done, _ = env.step(action=ac)
            data.append(state=st,
                        new_state=new_st,
                        reward=re,
                        action=ac,
                        done=done)
        print(algo.train(batch_data=data))

    def test_init_continuous(self):
        algo, locals = self.create_mpc(env_id='Pendulum-v0')
        env_spec = locals['env_spec']
        env = locals['env']
        algo.init()
        for _ in range(100):
            assert env_spec.action_space.contains(algo.predict(env_spec.obs_space.sample()))

        st = env.reset()
        data = TransitionData(env_spec)

        for _ in range(10):
            ac = algo.predict(st)
            new_st, re, done, _ = env.step(action=ac)
            data.append(state=st,
                        new_state=new_st,
                        reward=re,
                        action=ac,
                        done=done)
        print(algo.train(batch_data=data))

    def test_mpc_polymorphism(self):
        policy_func = (
            self.create_mlp_deterministic_policy, self.create_normal_dist_mlp_policy, self.create_uniform_policy,
            self.create_constant_action_policy)
        for i, func in enumerate(policy_func):
            self.setUp()
            wrap_policy(self, func=func)()
            if i < len(policy_func) - 1:
                self.tearDown()


def wrap_policy(self, func):
    def wrap_func():
        mlp_dyna, local = self.create_continue_dynamics_model(env_id='Pendulum-v0')
        env_spec = local['env_spec']
        env = local['env']
        policy = func(env_spec=env_spec)[0]
        algo, locals = self.create_mpc(env_spec=env_spec, mlp_dyna=mlp_dyna, policy=policy, env=env)
        algo.init()
        for _ in range(100):
            assert env_spec.action_space.contains(algo.predict(env_spec.obs_space.sample()))

        st = env.reset()
        data = TransitionData(env_spec)

        for _ in range(10):
            ac = algo.predict(st)
            new_st, re, done, _ = env.step(action=ac)
            data.append(state=st,
                        new_state=new_st,
                        reward=re,
                        action=ac,
                        done=done)
        print(algo.train(batch_data=data))

    return wrap_func


if __name__ == '__main__':
    unittest.main()
